use strict;
use warnings;

{
  package File::3DS;
  
  my %chunks = ( # Limited list - only those that I ever thought I might use
    to_id => {
      MAIN3DS                 => 0x4D4D,
       EDIT3DS                => 0x3D3D,
        EDIT_OBJECT           => 0x4000,
         OBJ_TRIMESH          => 0x4100,
          TRI_VERTEXL         => 0x4110,
          TRI_VERTEXOPTIONS   => 0x4111,
          TRI_MAPPINGCOORDS   => 0x4140,
          TRI_MAPPINGSTANDARD => 0x4170,
          TRI_FACEL1          => 0x4120,
           TRI_SMOOTH         => 0x4150,
           TRI_MATERIAL       => 0x4130,
       EDIT_MATERIAL          => 0xAFFF,
        MAT_NAME01            => 0xA000,
        MAT_DIFFUSE           => 0xA020,
         COL_RGBBYTE          => 0x0011,
        MAT_TEXTURE01         => 0xA200,
        MAT_OPACITY           => 0xA210,
         MAP_FILENAME         => 0xA300,
    },
    from_id => {},
  );
  while (my ($key, $value) = each %{$chunks{to_id}}) {
    $chunks{from_id}{$value} = $key;
  }
  
  sub new {
    my $self = shift;
    my %params = @_;
    return File::3DS::Chunk->new(
      type => 'MAIN3DS',
      data => {},
      subchunks => [],
      %params,
    );
  }
  
  sub chunk_id {
    my $self = shift;
    my $chunk = ref($self)?shift:$self;
    if (not defined $chunks{to_id}{$chunk}) {
      die "Invalid chunk $chunk!";
    }
    return $chunks{to_id}{$chunk};
  }
}

#### #### #### ####

#### #### #### ####

{
  package File::3DS::Chunk; # From which all specific chunks inherit
  
  sub new {
    my $chunk = {};
    my $self = shift;
    if (ref($self)) {
      $self->add($chunk);
    }
    my %params = @_;
    $chunk->{$_} = $params{$_} foreach (qw(type data subchunks));
    bless $chunk, 'File::3DS::Chunk::'.$chunk->{type};
    return $chunk;
  }
  
  sub add {
    my $self = shift;
    my $chunk = shift;
    push @{$self->{subchunks}}, $chunk;
  }
  
  sub write {
    my $self = shift;
    my %params = @_;
    my $contents = $self->write_contents(%params);
    return pack(SL => File::3DS::chunk_id($self->{type}), length($contents)+6) . $contents;
  }
  
  sub write_chunks {
    my $self = shift;
    my $chunks = shift;
    my %params = @_;
    my $return = '';
    $return .= $_->write(%params) foreach @$chunks;
    return $return;
  }
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return $self->write_chunks($self->{subchunks}, %params);
  }
}

#### #### #### ####
{
  package File::3DS::Chunk::MAIN3DS;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::EDIT3DS;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::EDIT_OBJECT;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack('Z*' => $self->{data}{name})
         . $self->write_chunks($self->{subchunks}, %params);
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::OBJ_TRIMESH;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::TRI_VERTEXL;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack(S => $#{$self->{data}{vertices}}+1) .
           join('', map pack(F => $_),
                    map $params{scale}?$_*$params{scale}:$_,
                    map @$_, @{$self->{data}{vertices}});
  }
}
#### #### #### ####
# {
  # package File::3DS::Chunk::TRI_VERTEXOPTIONS;
  # use base 'File::3DS::Chunk';
  #  (fairly useless)
# }
#### #### #### ####
{
  package File::3DS::Chunk::TRI_MAPPINGCOORDS;
  use base 'File::3DS::Chunk';
  # short: num verts
  # floats: U and V coords per vertex
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack(S => scalar(@{$self->{data}{coords}}))
         . join('', map pack(F => $_), map @$_, @{$self->{data}{coords}});
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::TRI_MAPPINGSTANDARD;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::TRI_FACEL1;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack(S => $#{$self->{data}{faces}}+1)
         . join('', map pack(S => $_), map @$_, @{$self->{data}{faces}})
         . $self->write_chunks($self->{subchunks}, %params);
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::TRI_SMOOTH;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return join '', @{$self->{data}{faces}};
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::TRI_MATERIAL;
  use base 'File::3DS::Chunk';
  # one for every material used
  # asciiz: name
  # short: num faces with this mtrl
  # shorts: list of faces
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack('Z*' => $self->{data}{name})
         . pack(S => scalar(@{$self->{data}{faces}}))
         . join('' => map pack(S => $_), @{$self->{data}{faces}});
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::EDIT_MATERIAL;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::MAT_NAME01;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack('Z*' => $self->{data}{name});
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::MAT_DIFFUSE;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::COL_RGBBYTE;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack(CCC => $self->{data}{r}, $self->{data}{g}, $self->{data}{b});
  }
}
#### #### #### ####
{
  package File::3DS::Chunk::MAT_TEXTURE01;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::MAT_OPACITY;
  use base 'File::3DS::Chunk';
}
#### #### #### ####
{
  package File::3DS::Chunk::MAP_FILENAME;
  use base 'File::3DS::Chunk';
  
  sub write_contents {
    my $self = shift;
    my %params = @_;
    return pack('Z*' => $self->{data}{filename});
  }
}
#### #### #### ####

1;
