use strict;
use warnings;

use Convert;

use constant DEV_VERSION => 0;
$main::autoflush = 1 if DEV_VERSION;

use Tk;
use Tk::Font;
use Tk::ROText;
use Tk::Balloon;
use Tk::ErrorDialog;

use Win32::API;
use Win32::FileOp;
use Win32::Registry;
use BrowseForFolder;

use File::Find();

use constant RGN_AND => 1;
use constant RGN_OR  => 2;
use constant RGN_XOR => 3;
use constant RGN_DIFF=> 4;
use constant RGN_COPY=> 5;
use constant AW_HIDE    => 0x10000;
use constant AW_ACTIVATE=> 0x20000;
use constant AW_BLEND   => 0x80000;


my %skin_settings;
foreach (maybe_bound_filedata('aom.skn')) {
  s/\r?\n$//g; # because chomp doesn't like \r\n
  next if /^#/;
  next unless $_;
  my ($key, $value) = split(/:\s+/, $_, 2);
  $skin_settings{$key} = $value;
}

my %fonts;


my ($primary_window, $main_window);
$primary_window = MainWindow->new( # so that there's always a selectable window for alt-tab/taskbar/etc
  -background => '#000000',
  -title => $skin_settings{title},
  -width => 1,
  -height => 1,
);
$primary_window->iconify;

my $icon = $primary_window->Photo(-file => maybe_bound_filename($skin_settings{icon_image}));

my $CreateRectRgn  =   new Win32::API(gdi32 => CreateRectRgn      => [qw(I I I I)],     'N');
my $CreateRoundRectRgn=new Win32::API(gdi32 => CreateRoundRectRgn => [qw(I I I I I I)], 'N');
my $CombineRgn     =   new Win32::API(gdi32 => CombineRgn         => [qw(N N N I)],     'N');
my $DeleteObject   =   new Win32::API(gdi32 => DeleteObject       => ['N'],             'N');
my $SetWindowRgn   =   new Win32::API(user32=> SetWindowRgn       => [qw(N N N)],       'N');

my $help_file;
sub help_filename {
  return $help_file if $help_file;
  return maybe_bound_filename("help.htm");
}

my @known_file_extensions = qw(bar ddt brg scn lib xmb bdt btx bt8 rec trg drs);
my $file_selection_filter = [
  [ "Known data files (".join(', ', @known_file_extensions).")", join(';', map "*.$_", @known_file_extensions) ],
  [ "All files", "*.*" ],
];

my @ddt_types = (
  '003' => 'Paletted (16-bit), no alpha [0]',
  '203' => 'Paletted (16-bit), no alpha [2]',
  '303' => 'Paletted (16-bit), no alpha [3]',
  '403' => 'Paletted (16-bit), no alpha [4]',
  '213' => 'Paletted (15-bit), no alpha',
  '243' => 'Paletted (12-bit), 4-bit alpha',
  '004' => '16-bit, no alpha [0]',
  '104' => '16-bit, no alpha [1]',
  '204' => '16-bit, no alpha [2]',
  '804' => '16-bit, no alpha [8]',
  '015' => '15-bit, 1-bit alpha [0]',
  '115' => '15-bit, 1-bit alpha [1]',
  '215' => '15-bit, 1-bit alpha [2]',
  '315' => '15-bit, 1-bit alpha [3]',
  '046' => '12-bit, 4-bit alpha [0]',
  '146' => '12-bit, 4-bit alpha [1]',
  '246' => '12-bit, 4-bit alpha [2]',
  '087' => '8-bit greyscale, no alpha',
);

my @simple_file_extensions = qw(xmb xml ddt bmp brg 3dx);
my $simple_file_selection_filter = [
  [ "Known data files (".join(', ', @simple_file_extensions).")", join(';', map "*.$_", @simple_file_extensions) ],
  [ "All files", "*.*" ],
];

my %simple_conversions = (
  # in format  extension => [ [ new extension => sub for conversion, sub for param generation, sub for finishing ] ]

  xml => [ [ xmb => \&Convert::xml_to_xmb   ] ],

  xmb => [ [ xml => \&Convert::xmb_to_xml   ] ],

  ddt => [ [ bmp => \&Convert::ddt_to_image, undef, sub {
    my $return = $_[0];
    my %ddt_types = @ddt_types;
    $main_window->messageBox(
      -message => qq{DDT image format: "$ddt_types{$return->{type}}", $return->{levels} mip-map level}.($return->{levels}==1?'':'s').qq{ (remember this if you want to recreate the DDT!)},
      -title => 'Completed',
      -type => 'OK');
    return 1;
  } ] ],

  bmp => [ [ ddt => \&Convert::image_to_ddt, sub {
    my $format_dialog = $main_window->DialogBox(-title => 'DDT format', -buttons => ['OK','Cancel']);

    my $radio_format = '003';
    my @radio_format = @ddt_types;

    my ($k, $v);
    while (($k, $v) = (shift @radio_format, shift @radio_format), defined $k) {
      $format_dialog->add(Radiobutton =>
        -variable => \$radio_format,
        -value => $k,
        -text => $v,
        -selectcolor => '#ffffff',
      )->pack(
        -fill => 'x',
        -anchor => 'w',
      );
    }
    return undef if $format_dialog->Show ne 'OK';

    my ($radio_levels, @radio_levels);
    if ($radio_format =~ /0.3/) {
      @radio_levels = (2,5);
      $radio_levels = 2;
    } elsif ($radio_format =~ /[^0].3/) {
      @radio_levels = (1);
      $radio_levels = 1;
    } elsif ($radio_format =~ /..4/) {
      @radio_levels = (1,4,5);
      $radio_levels = 5;
    } elsif ($radio_format =~ /..5/) {
      @radio_levels = (1,4,5);
      $radio_levels = 5;
    } elsif ($radio_format =~ /..6/) {
      @radio_levels = (1,3,4,5);
      $radio_levels = 5;
    } elsif ($radio_format =~ /..7/) {
      @radio_levels = (2);
      $radio_levels = 2;
    } else {
      @radio_levels = (1,2,3,4,5);
      $radio_levels = 5;
    }

    if (@radio_levels > 1) {
      my $levels_dialog = $main_window->DialogBox(-title => 'Mipmap levels', -buttons => ['OK','Cancel']);
      foreach (@radio_levels) {
        $levels_dialog->add(Radiobutton =>
          -variable => \$radio_levels,
          -value => "$_",
          -text => "$_ levels",
          -selectcolor => '#ffffff',
        )->pack(
          -fill => 'x',
          -anchor => 'w',
        );
      }
      return undef if $levels_dialog->Show ne 'OK';
    }

    return { format => $radio_format, levels => $radio_levels };
  } ] ],

  brg => [ [ '3ds' => \&Convert::brg_to_3ds, sub {
    my $bar_dialog = $main_window->DialogBox(-title => 'Textures.bar location', -buttons => ['OK','Cancel']);

    $bar_dialog->Label(-text => 'Textures.bar is required for automatic texture conversion. Choosing "cancel" will disable this feature.')->pack();

    my $texturesbar_location = File::Spec->canonpath(aom_AppPath('/textures/textures.bar') . '/textures/textures.bar');

    $bar_dialog->Button(
      -font => $fonts{button},
      -text => 'Browse for file',
      -command => sub {
        my $filename = $main_window->getOpenFile(
            -title => 'Select textures.bar',
            -filetypes => [ [ "BAR files (*.bar)", "*.bar" ], [ "All files (*.*)", "*.*" ] ],
            -initialdir => to_path($texturesbar_location),
            -initialfile => $texturesbar_location,
        );
        $filename = File::Spec->canonpath($filename);
        $texturesbar_location = $filename if $filename and -e $filename;
      },
    )->pack(-padx => 2, -side => 'left');

    $bar_dialog->Entry(
      -font => $fonts{entry},
      -textvariable => \$texturesbar_location,
    )->pack(-padx => 2, -side => 'left', -fill => 'x', -expand => 1);

    $texturesbar_location = undef if $bar_dialog->Show ne 'OK';

    return { texturesbar => $texturesbar_location };
  }, sub {
    my $return = $_[0];
    $main_window->messageBox(
      -message => 'Conversion finished. To import the 3DS into 3ds max or gmax with proper animation and textures, use the brg_import.ms script (available somewhere online - try mods.dgdn.net)',
      -title => 'Completed',
      -type => 'OK');
    return 1;
  } ] ],

  '3dx'=> [ [ 'brg' => \&Convert::threeds_to_brg, sub {
    my $brg_dialog = $main_window->DialogBox(-title => 'Base BRG location', -buttons => ['OK','Cancel']);

    $brg_dialog->Label(-text => "BRG creation currently must be based on an existing BRG with the\nsame length animation and materials. Please choose an appropriate BRG:")->pack();

    my $basebrg_location = File::Spec->canonpath(aom_AppPath().'/');

    $brg_dialog->Button(
      -font => $fonts{button},
      -text => 'Browse for file',
      -command => sub {
        my $filename = $main_window->getOpenFile(
            -title => 'Select base BRG',
            -filetypes => [ [ "BRG files (*.brg)", "*.brg" ], [ "All files (*.*)", "*.*" ] ],
            -initialdir => to_path($basebrg_location),
            -initialfile => $basebrg_location,
        );
        $filename = File::Spec->canonpath($filename);
        $basebrg_location = $filename if $filename and -e $filename;
      },
    )->pack(-padx => 2, -side => 'left');

    $brg_dialog->Entry(
      -font => $fonts{entry},
      -textvariable => \$basebrg_location,
    )->pack(-padx => 2, -side => 'left', -fill => 'x', -expand => 1);

    return undef if $brg_dialog->Show ne 'OK';

    return { brg => $basebrg_location };
  } ] ],
);

my $program_path;
if ($PerlApp::BUILD) {
  $program_path = to_path(PerlApp::exe());
} else {
  my $require = 'FindBin.pm'; # trick PerlApp into not including it (for no particular reason)
  require $require;
  $program_path = File::Spec->canonpath($FindBin::Bin);
}

$main_window = $primary_window->Toplevel(
  -title => $skin_settings{title},
  -width => (split /\s+/, $skin_settings{main_window_size})[0],
  -height => (split /\s+/, $skin_settings{main_window_size})[1],
);
$main_window->iconify;
$main_window->resizable(0,0);
$main_window->overrideredirect(1);
$main_window->geometry('+100+100'); # set an initial position

$primary_window->bind('<Visibility>', sub { $primary_window->iconify; $main_window->focusForce });

my $balloon = $primary_window->Balloon(
  -state => 'balloon',
);

my $bg_image = $main_window->Label(
  -borderwidth => 0,
  -image => $primary_window->Photo(
    -file => maybe_bound_filename($skin_settings{background_image}),
  ),
)->place(
  -x => 0,
  -y => 0,
);
$balloon->attach(
  $bg_image,
  -balloonmsg => 'Age of Mythology Data File Converter - created by Ykkrosh (ykkrosh@zaynar.demon.co.uk)',
  -initwait => 1000,
  -balloonposition => 'mouse',
);

sub extract_source {
  my $directory = BrowseForFolder_(
    title  => 'Extract source code',
    text   => q{Select a folder to extract the program's code into},
    folder => $program_path,
  );
  if ($directory and -d $directory) {
    $directory = File::Spec->canonpath($directory);
    foreach my $file (qw(fe.pl TkGUI.pm Convert.pm ThreeDS.pm Triggers.pm triggerlist.pl resrip.c resrip.i BrowseForFolder.pm help.htm aom_back.gif aomicon.gif aom.skn default.fmt readme.txt)) {
      open my $fh, ">$directory\\$file" or die "Error opening $directory\\$file for output: $!";
      binmode $fh;
      print $fh (maybe_bound_filedata($file));
      close $fh;
    }
  }
}
$bg_image->bind('<Control-Shift-Alt-Quadruple-Button-1>', \&extract_source);

sub extract_maxscripts {
  $main_window->messageBox(
    -message => 'These are probably not the latest versions of the scripts - check mods.dgdn.net for information on any updates',
    -title => 'Extract BRG MAXScripts',
    -type => 'OK');
  my $directory = BrowseForFolder_(
    title  => 'Extract BRG MAXScripts',
    text   => q{Select a folder to extract the .ms files into},
    folder => $program_path,
  );
  if ($directory and -d $directory) {
    $directory = File::Spec->canonpath($directory);
    foreach my $file (qw(brg_import.ms brg_export.ms)) {
      open my $fh, ">$directory\\$file" or die "Error opening $directory\\$file for output: $!";
      #binmode $fh;
      print $fh (maybe_bound_filedata($file));
      close $fh;
    }
  }
}

{
  my %palette;
  foreach (qw(background foreground selectBackground selectForeground)) {
    $palette{$_} = $skin_settings{$_} if exists $skin_settings{$_};
  }
  $main_window->setPalette(%palette);
}

{
  my $main_window_drag;
  
  $main_window->bind('<ButtonPress-1>' => [ sub {
    if ($_[0] eq $bg_image) {
      $main_window_drag = [$main_window->rootx - $_[1], $main_window->rooty - $_[2]];
    }
   }, Ev('X'), Ev('Y')
  ]);
  
  $main_window->bind('<ButtonRelease-1>' => sub { undef $main_window_drag });
  
  $main_window->bind('<Motion>' => [ sub {
    if ($main_window_drag) {
      $main_window->geometry('+'.($_[1]+$main_window_drag->[0]).'+'.($_[2]+$main_window_drag->[1]));
    }
  }, Ev('X'), Ev('Y')
  ]);
}


foreach (qw(command_button button entry log)) {
  my @font_data = split(/\s+/, $skin_settings{"${_}_font"});
  my %font_data = (
    -family => shift @font_data,
    -size => -(shift @font_data),
  );
  $font_data{slant} = 'italic' if grep $_ eq 'italic', @font_data;
  $font_data{weight} = 'bold' if grep $_ eq 'bold', @font_data;

  $fonts{$_} = $main_window->Font(%font_data);
}


my $help_button = $main_window->Button(
  -text => '?',
  -font => $fonts{command_button},
  -command => sub {
    my $return = Win32::FileOp::ShellExecute(open => help_filename());

#    if ($return <= 32) { # doesn't work!!
#      if (system(start => help_filename())) {
#        die "Error opening the HTML help file ($?)! Make sure you have a browser installed";
#      }
#    }
  },
)->place(
  -x => (split /\s+/, $skin_settings{help_button_pos})[0],
  -y => (split /\s+/, $skin_settings{help_button_pos})[1],
  -width => (split /\s+/, $skin_settings{command_button_size})[0],
  -height => (split /\s+/, $skin_settings{command_button_size})[1],
);
my $exit_button = $main_window->Button(
  -text => 'X',
  -font => $fonts{command_button},
  -command => [ $primary_window => 'destroy' ],
)->place(
  -x => (split /\s+/, $skin_settings{exit_button_pos})[0],
  -y => (split /\s+/, $skin_settings{exit_button_pos})[1],
  -width => (split /\s+/, $skin_settings{command_button_size})[0],
  -height => (split /\s+/, $skin_settings{command_button_size})[1],
);

$balloon->attach($help_button, -balloonmsg => 'Open help file');
$balloon->attach($exit_button, -balloonmsg => 'Exit');

my @gui_sections = (
  [ set_read_file    => ['Set input data file', 'Select a data file which the program will read and convert into editable text files', \&set_read_file],
    read_file_label  => \$main::settings->{input_file}],
  [ set_read_dir     => ['Folder to extract data into', 'Select a folder into which the editable text files will be placed (be aware that it will overwrite existing text data)', \&set_read_dir], 
    read_dir_label   => \$main::settings->{output_text_path}],
  [ do_read_button   => ['Read data file', '', \&do_read_button] ],
  undef,
  [ set_write_file   => ['Set output data file', 'Select a data file which the program will create, based on the specified text files', \&set_write_file],
    write_file_label => \$main::settings->{output_file}],
  [ set_write_dir    => ['Folder to read data from', 'Select a folder to read the text data from', \&set_write_dir],
    write_dir_label  => \$main::settings->{input_text_path}],
  [ do_write_button  => ['Write data file', '', \&do_write_button] ],
);

my @gui_entries;
my @gui_entry_widgets;

{
  my $vpos = (split /\s+/, $skin_settings{button_pos})[1];
  for my $set (0..$#gui_sections) {
    if (not defined $gui_sections[$set]) {
      $vpos += $skin_settings{button_empty_spacing};
      next;
    }
    if ($gui_sections[$set][0]) {
      my $button = $main_window->Button(
        -font => $fonts{button},
        -text => $gui_sections[$set][1][0],
        -command => sub { validate_entries(); $gui_sections[$set][1][2]->() },
      )->place(
        -x => (split /\s+/, $skin_settings{button_pos})[0],
        -y => $vpos,
        -width => (split /\s+/, $skin_settings{button_size})[0],
        -height => (split /\s+/, $skin_settings{button_size})[1],
      );
      if ($gui_sections[$set][1][0]) {
        $balloon->attach($button, -balloonmsg => $gui_sections[$set][1][1]);
      }
    }
    if ($gui_sections[$set][2]) {
      my $id = $#gui_entries + 1;
      $gui_entries[$id] = $gui_sections[$set][3];
      ${$gui_entries[$id]} = '' unless defined ${$gui_entries[$id]};

      $gui_entry_widgets[$id] = $main_window->Entry(
        -font => $fonts{entry},
        -textvariable => $gui_entries[$id],
      );
      $gui_entry_widgets[$id]->bind('<FocusOut>', \&validate_entries);
      $gui_entry_widgets[$id]->place(
        -x => (split /\s+/, $skin_settings{entry_pos})[0],
        -y => $vpos - (split /\s+/, $skin_settings{button_pos})[1]+(split /\s+/, $skin_settings{entry_pos})[1],
        -width => (split /\s+/, $skin_settings{entry_size})[0],
        -height => (split /\s+/, $skin_settings{entry_size})[1],
      );
    }
    $vpos += $skin_settings{button_spacing};
  }
}

{
  my $button = $main_window->Button(
    -font => $fonts{button},
    -text => 'Direct file conversion',
    -command => sub { simple_convert() },
  )->place(
    -x => (split /\s+/, $skin_settings{simple_button_pos})[0],
    -y => (split /\s+/, $skin_settings{simple_button_pos})[1],
    -width => (split /\s+/, $skin_settings{simple_button_size})[0],
    -height => (split /\s+/, $skin_settings{simple_button_size})[1],
  );
  $balloon->attach($button, -balloonmsg => 'Perform a simple conversion of a file into another type');
}
{
  my $button = $main_window->Button(
    -font => $fonts{button},
    -text => 'Bulk conversion',
    -command => sub { bulk_convert() },
  )->place(
    -x => (split /\s+/, $skin_settings{bulk_button_pos})[0],
    -y => (split /\s+/, $skin_settings{bulk_button_pos})[1],
    -width => (split /\s+/, $skin_settings{bulk_button_size})[0],
    -height => (split /\s+/, $skin_settings{bulk_button_size})[1],
  );
  $balloon->attach($button, -balloonmsg => 'Convert all the files in a folder at once');
}
{
  my $button = $main_window->Button(
    -font => $fonts{button},
    -text => 'Advanced options',
    -command => sub { advanced_options() },
  )->place(
    -x => (split /\s+/, $skin_settings{advanced_button_pos})[0],
    -y => (split /\s+/, $skin_settings{advanced_button_pos})[1],
    -width => (split /\s+/, $skin_settings{advanced_button_size})[0],
    -height => (split /\s+/, $skin_settings{advanced_button_size})[1],
  );
  $balloon->attach($button, -balloonmsg => 'Miscellaneous possibilities that aren\'t generally useful');
}

our %advanced_flags = (
  scn_unit_edit => 0,
  scn_text_trigger => 0,
);


my ($log_window, $log_window_text);

#create_log_window();

sub create_log_window {
  $log_window = $primary_window->Toplevel(
    -title => 'Log window',
  );
  $log_window->bind('<Destroy>', sub { undef $log_window; undef $log_window_text });
  $log_window_text = $log_window->Scrolled(
    ROText => -scrollbars => 'e',
    -font => $fonts{log},
    -background => '#000000',
    -foreground => '#00ff00',
    -width => 45,
    -height => 25,
  )->pack(
    -expand => 1,
    -fill => 'both',
  );

  $log_window->update;
  $log_window->iconimage($icon);
}

$SIG{__WARN__} = \&log_add;

sub start_gui {

#  log_clear();
#  log_add('Initialising');

  # Do the curvey edges
  my $main_region = $CreateRoundRectRgn->Call(0, 0, (split /\s+/, $skin_settings{main_window_size})[0]+1, (split /\s+/, $skin_settings{main_window_size})[1]+1, 100, 100);
  $main_window->update; # need to show the window before ->frame will work
  $SetWindowRgn->Call(hex($main_window->frame), $main_region, 1);
  $main_window->deiconify;

  $primary_window->update;
  $primary_window->iconimage($icon);

  read_format_file();

  if ($skin_settings{startup_warning}) {
    $main_window->messageBox(
      -message => $skin_settings{startup_warning},
      -title => 'Warning',
      -type => 'OK');
  }

  MainLoop;

  $DeleteObject->Call($main_region) if defined $main_region;

  exit;
}

sub read_format_file {
  my @raw_format;
  if ($PerlApp::BUILD) {
    @raw_format = PerlApp::get_bound_file("default.fmt") or die "Error reading bound file default.fmt!";
  } else {
    open FMT, "default.fmt" or die "Error opening default.fmt: $!";
    @raw_format = <FMT>;
    close FMT;
  }
  $main::format = parse_format(@raw_format);
}

my ($simple_convert_start_dir, $simple_convert_start_file);

sub simple_convert {
  $simple_convert_start_dir ||= to_path($main::settings->{input_file});
  $simple_convert_start_file ||= $main::settings->{input_file} || '';
  $simple_convert_start_file .= '*.*' if substr($simple_convert_start_file, -1) eq '\\';
  my $input_filename = $main_window->getOpenFile(
      -title => 'Select data file for input',
      -filetypes => $simple_file_selection_filter,
      -initialdir => $simple_convert_start_dir,
      -initialfile => $simple_convert_start_file,
  );
  $input_filename = File::Spec->canonpath($input_filename);
  if (not $input_filename or not -e $input_filename) {
    return 1;
  }
  my $file_extn = lc(($input_filename =~ /.*\.(.+)$/)[0]);

  my $output_type = $simple_conversions{$file_extn}[0];

  if (not $output_type) {
    $main_window->messageBox(
      -message => "No known conversion for filetype '$file_extn'!",
      -title => 'Error',
      -type => 'OK');
    return 1;
  }

  $main_window->messageBox(
    -message => 'Converting to '.uc($output_type->[0]).' - select an output file in the following window',
    -title => 'Converting',
    -type => 'OK');

  $simple_convert_start_dir = to_path($input_filename);
  $simple_convert_start_file = $input_filename;
  my $start_file = ($input_filename =~ /^(.*)\..+$/)[0] . ".$output_type->[0]";
  my $output_filename = $main_window->getSaveFile(
      -title => 'Select new data file to create',
      -filetypes => [ [ uc($output_type->[0]), "*.$output_type->[0]" ],
                      [ "All files", "*.*" ]
                    ],
      -initialdir => $simple_convert_start_dir,
      -initialfile => $start_file,
  );
  $output_filename = File::Spec->canonpath($output_filename);

  if (not $output_filename) {
    return 1;
  }

  my $params = {};
  if ($output_type->[2]) {
    $params = $output_type->[2]->($input_filename, $output_filename);
    return 1 unless defined $params;
  }

  my $return = $output_type->[1]->($input_filename, $output_filename, $params);

  if ($return and $output_type->[3]) {
    return if $output_type->[3]->($return);
  }

  $main_window->messageBox(
    -message => 'Conversion finished',
    -title => 'Completed',
    -type => 'OK');
}

sub bulk_convert {

  my $start_dir = to_path($main::settings->{input_file});
  my $input_dir = BrowseForFolder_(
    title  => 'Bulk conversion',
    text   => 'Select the folder containing the files you wish to convert (files in sub-folders will also be converted)',
    folder => $start_dir,
  );
  $input_dir = File::Spec->canonpath($input_dir);
  if (not $input_dir or not -d $input_dir) {
    return 1;
  }

  my %file_types;

  File::Find::find( { wanted => sub {
    foreach (keys %simple_conversions) {
      if (lc(substr($File::Find::name, -length($_))) eq $_) { $file_types{$_} ||= []; push @{$file_types{$_}}, $File::Find::name; return; }
    }
  }, no_chdir => 1 }, $input_dir);

  if (scalar keys %file_types == 0) {
    $main_window->messageBox(
      -message => 'No convertable files have been found!',
      -title => 'Unable to continue',
      -type => 'OK');
    return 1;
  }

  return 1 if $main_window->messageBox(
      -message => 'Found ' . nice_join(map scalar(@{$file_types{$_}}).' '.uc($_).' file'.($#{$file_types{$_}}?'s':''), sort { $#{$file_types{$a}} <=> $#{$file_types{$b}} } keys %file_types)  # erm...?!
                .'. Continue?',
      -title => 'Files found',
      -type => 'OKCancel') ne 'ok';

  my $output_dir = BrowseForFolder_(
    title  => 'Bulk conversion',
    text   => 'Select the folder to put the converted files into (preferably somewhere empty)',
    folder => $input_dir,
  );
  $output_dir = File::Spec->canonpath($output_dir);
  if (not $output_dir or not -d $output_dir) {
    return 1;
  }

  my $num_files = 0; $num_files += scalar @$_ foreach values %file_types;
  return 1 if $main_window->messageBox(
      -message => "Continue conversion of $num_files file" . ($num_files==1?'':'s') . '?' . ($num_files > 100 ? ' (this may take some time)' : ''),
      -title => 'Confirmation',
      -type => 'OKCancel') ne 'ok';

  my $overwrite_alert = 0;
  foreach my $in_file (sort map @$_, values %file_types) {
    $in_file = File::Spec->canonpath($in_file);
    $in_file =~ /^\Q$input_dir\E\\(.*)$/i or die "File '$in_file' doesn't appear to be in '$input_dir'!";
    my $out_file = "$output_dir\\$1";
    my $file_extn = lc(($in_file =~ /.*\.(.+)$/)[0]);
    my $output_type = $simple_conversions{$file_extn}[0];
    $out_file = ($out_file =~ /^(.*)\..+$/)[0] . ".$output_type->[0]";

    log_add("$in_file --> $out_file");

    my @paths = File::Spec->splitdir($out_file);
    for (0..$#paths-1) {
      mkdir join('/', @paths[0..$_]) or die "Error creating ".join('/', @paths[0..$_]).": $!" unless -e join('/', @paths[0..$_]);
    }
    $output_type->[1]->($in_file, $out_file);
  }

  $main_window->messageBox(
    -message => 'Conversions finished',
    -title => 'Completed',
    -type => 'OK');
}


sub advanced_options {
  my $dialog = $main_window->DialogBox(
    -title => 'Advanced options',
    -background => $skin_settings{dialogbackground},
    -buttons => ['Close']);

  $dialog->Label(
    -text => "If you don't know what these options are for, you\nprobably don't need to, so don't worry about them ;-)",
    -background => $skin_settings{dialogbackground},
  )->pack();

  $dialog->Button(
    -text => 'Extract BRG MAXScripts',
    -command => \&extract_maxscripts,
  )->pack(-anchor => 'w');

  $dialog->Button(
    -text => 'Extract source code',
    -command => \&extract_source,
  )->pack(-anchor => 'w');

  my @check_optns = (
    -onvalue => 1,
    -offvalue => 0,
    -selectcolor => '#ffffee',
    -background => $skin_settings{dialogbackground},
  );

  $dialog->Checkbutton( @check_optns,
    -text => 'Enable SCN unit editing',
    -variable => \$advanced_flags{scn_unit_edit},
  )->pack(-anchor => 'w');

  $dialog->Checkbutton( @check_optns,
    -text => 'Export SCN triggers as text',
    -variable => \$advanced_flags{scn_text_trigger},
  )->pack(-anchor => 'w');

  $dialog->Show();

}


sub set_read_file {
  my $start_dir = to_path($main::settings->{input_file});
  my $start_file = $main::settings->{input_file} || '';
  $start_file .= '*.*' if substr($start_file, -1) eq '\\';
  my $filename = $main_window->getOpenFile(
      -title => 'Select data file for input',
      -filetypes => $file_selection_filter,
      -initialdir => $start_dir,
      -initialfile => $start_file,
  );
  $filename = File::Spec->canonpath($filename);
  if ($filename and -e $filename) {
    ${$gui_entries[0]} = $filename;
    validate_entries();
  }
}

sub set_read_dir {
  my $start_dir = $main::settings->{output_text_path} || $program_path;
  my $directory = BrowseForFolder_(
    title  => 'Extract text data',
    text   => 'Select a folder to extract the text data into',
    folder => $start_dir,
  );
  if ($directory and -d $directory) {
    $directory = File::Spec->canonpath($directory);
    ${$gui_entries[1]} = $directory;
    validate_entries();
  }
}

sub do_read_button {
  if (not $main::settings->{input_file} or not $main::settings->{output_text_path}) {
    $main_window->messageBox(
      -message => 'You must select a file to read and a folder to store the data in!',
      -title => 'Error',
      -type => 'OK');
    return 1;
  }

  if (not -e $main::settings->{input_file} or not -f $main::settings->{input_file}) {
    $main_window->messageBox(
      -message => 'The chosen input data file does not exist or is not a valid file!',
      -title => 'Error',
      -type => 'OK');
    return 1;
  }

  log_clear();

  read_format_file() if DEV_VERSION;

  log_add("Reading $main::settings->{input_file}");
  read_using($main::format);
  log_add("Finished!");
}

sub set_write_file {
  my $start_dir = to_path($main::settings->{output_file});
  my $start_file = $main::settings->{output_file} || '';
  $start_file .= '*.*' if substr($start_file, -1) eq '\\';
  my $filename = $main_window->getSaveFile(
      -title => 'Select new data file to create',
      -filetypes => $file_selection_filter,
      -initialdir => $start_dir,
      -initialfile => $start_file,
  );
  $filename = File::Spec->canonpath($filename);
  if ($filename) {
    ${$gui_entries[2]} = $filename;
    validate_entries();
  }
}

sub set_write_dir {
  my $start_dir = $main::settings->{input_text_path} || $main::settings->{output_text_path} || $program_path;
  my $directory = BrowseForFolder_(
    title  => 'Read text data',
    text   => 'Select a folder to read the text data from',
    folder => $start_dir,
  );
  if ($directory and -d $directory) {
    $directory = File::Spec->canonpath($directory);
    ${$gui_entries[3]} = $directory;
    validate_entries();
  }
}

sub do_write_button {
  if (not $main::settings->{output_file} or not $main::settings->{input_text_path}) {
    $main_window->messageBox(
      -message => 'You must select a file to write and a folder to read the data from!',
      -title => 'Error',
      -type => 'OK');
    return 1;
  }

  if (-e $main::settings->{output_file}) {
    my $return = $main_window->messageBox(
      -message => "Overwrite file '$main::settings->{output_file}'?",
      -title => 'Confirm overwrite',
      -type => 'YesNo');
    return if $return eq 'no';
  }

  log_clear();
  log_add("Writing $main::settings->{output_file}");

  read_format_file() if DEV_VERSION;

  write_using($main::format);

  log_add("Finished!");
}


sub validate_entries {
  ${$gui_entries[1]} = to_path(${$gui_entries[0]}) if length ${$gui_entries[0]} and not length ${$gui_entries[1]};
  ${$gui_entries[2]} = ${$gui_entries[0]} if length ${$gui_entries[0]} and not length ${$gui_entries[2]};
  ${$gui_entries[3]} = ${$gui_entries[1]} if length ${$gui_entries[1]} and not length ${$gui_entries[3]};
  for my $id (0..$#gui_entries) {
    ${$gui_entries[$id]} = File::Spec->canonpath(${$gui_entries[$id]});
    $gui_entry_widgets[$id]->update;
  }
}

sub log_clear {
  create_log_window() if not defined $log_window_text;
  $log_window_text->delete('1.0', 'end');
  $log_window_text->insert('end', "LOG WINDOW\n==========\n");
}
sub log_add {
  create_log_window() if not defined $log_window_text;
  my ($message) = @_;
  $message .= "\n";
  $log_window_text->insert('end', $message);
  my $length = length($log_window_text->get('1.0','end'));
  if ($length > 8192) {
    $log_window_text->delete('1.0', '1.0 + ' . ($length-8192) . ' chars');
  }
  $log_window_text->see('end');

  $log_window->update;
  create_log_window() if not defined $log_window_text;
  $log_window->focus;
}

sub to_path {
  ($_[0]||'') =~ m#^(.*)[\\/].*$# or return File::Spec->canonpath($_[0]);
  return File::Spec->canonpath($1);
}

sub nice_join {
  return join '', map $_==0?$_[$_]:$_==$#_?" and $_[$_]":", $_[$_]", 0..$#_;
}

sub maybe_bound_filedata {
  my $filename = shift;
  if ($PerlApp::BUILD) {
    if (wantarray) {
      my @return = PerlApp::get_bound_file($filename);
      die "Unable to find bound file '$filename'!" unless @return;
      return @return;
    } else {
      my $return = PerlApp::get_bound_file($filename);
      die "Unable to find bound file '$filename'!" unless defined $return;
      return $return;
    }
  }
  open BFILE, "< $filename" or die "Error opening $filename for input: $!";
  binmode BFILE;
  if (wantarray) {
    my @return = <BFILE>;
    close BFILE;
    return @return;
  } else {
    local $/ = undef;
    my ($return) = <BFILE>;
    close BFILE;
    return $return;
  }
}

sub maybe_bound_filename {
  my $filename = shift;
  if ($PerlApp::BUILD) {
    my $return = PerlApp::extract_bound_file($filename);
    die "Unable to find bound file '$filename'!" unless defined $return;
    return $return;
  } else {
    return $filename;
  }
}

sub BrowseForFolder_ {
  my %params = @_;
  return BrowseForFolder::new(
   -owner => hex($main_window->frame),
   -initial => CSIDL_DESKTOP,
   -folder  => $params{folder},
   -title   => $params{title},
   -text    => $params{text},
   -status  => 1,
  )->show;
}

sub aom_AppPath {
  my $test_extn = shift;
  my ($key, $value, $type);
  return $value if $::HKEY_LOCAL_MACHINE->Open("Software\\Microsoft\\Microsoft Games\\Age of Mythology\\1.0", $key)
               and $key->QueryValueEx('AppPath', $type, $value)
               and (!$test_extn or -e "$value/$test_extn");
  return $value if $::HKEY_LOCAL_MACHINE->Open("Software\\Microsoft\\Microsoft Games\\Age of Mythology Demo\\1.0", $key)
               and $key->QueryValueEx('AppPath', $type, $value)
               and (!$test_extn or -e "$value/$test_extn");
  return '';
}

1;
