################################################
# FILE EDITOR (ooh, imaginative...)             #
 #  by Philip Taylor (philip@zaynar.demon.co.uk) #
  #################################################
   #  Built mainly for use with AOM, but also
    #  to make my life much easier with any
     #  other data files I ever feel like
      #  editing...

use strict;
use warnings;

our $settings;

use Safe;
use Carp;
use File::Spec;
use Time::Local;
use IO::Scalar;
use IO::File;
use Compress::Zlib(); # (but don't export anything)

use Convert;

use constant FIELD_PACK_FORMAT => 'S';
use constant FIELD_VAR_PACK_FORMAT => 'S';
# change to L if there are ever >64k iterations of a value (and to Q if there are >4 billion, and maybe w if there are >18 quintillion...)

# Debugging commands
use constant SAY_LINE => 0;
use constant DEBUG => 0;
use constant HEAVY_DEBUG => 0;

# Enables patch files
use constant ALLOW_PATCH => 0;

use constant FH_CACHE_LIMIT => 4; # max filehandles open simultaneously

our $autoflush;

#use noGUI2;
use TkGUI;

my %command_words = (
  in => {
    seek         => \&command_seek,
    file         => \&command_file_in,
    file_time    => \&command_filetime_in,
    repeat_alter => \&command_repeat_alter,
    var_set      => \&command_var_set,
    var_undef    => \&command_var_undef,
    die          => \&command_die,
    echo         => \&command_echo,
    subformat    => \&command_subformat_in,
    forget       => \&command_forget,
    note         => \&command_note_in,
    ensure       => \&command_ensure,
    set_palette  => \&command_set_palette,
  },
  out => {
    write        => \&command_write,
    file         => \&command_file_out,
    file_time    => \&command_filetime_out,
    repeat_alter => \&command_repeat_alter,
    var_set      => \&command_var_set,
    var_undef    => \&command_var_undef,
    alias        => \&command_alias,
    die          => \&command_die,
    echo         => \&command_echo,
    subformat    => \&command_subformat_out,
    forget       => \&command_forget,
    ensure       => \&command_ensure,
  }
);

my %database;

#my %fields = (from_id => [], to_id => {}, type => []);
my %fields;

my %aliases;

my $EOF_count = 0;

start_gui();




sub read_using {
  my $tree = shift;
  $settings->{parse_tree} = $tree;

  my $fh = IO::File->new("< $settings->{input_file}");
  if (not defined $fh) { die qq{Unable to open $settings->{input_file} for input: $!} };
  binmode $fh;
  $settings->{input_fh} = $fh;
  $settings->{input_fh_length} = (stat $settings->{input_file})[7];

  if ($tree->{type} ne 'branch' or $tree->{action} ne 'format_list') {
    die "Format file must begin with a format_list!\n";
  }
  $settings->{first_branch} = determine_format(@{$tree->{contents}});
  print "Using format $settings->{format_name}\n" if DEBUG;

  my @format_params = @{$settings->{first_branch}{params}};
  for my $pid (0..$#format_params) {
    if (defined $format_params[$pid] and $format_params[$pid] eq 'compression') {
      my $data_format = $format_params[$pid+1];

      my $new_data = decompress(join('', $fh->getlines), $data_format);
      $settings->{input_fh} = new IO::Scalar \$new_data;
      $settings->{input_fh_length} = length($new_data);

      splice @format_params, $pid, 2;
      redo;
    }
  }

  my $params = { action => 'read' };
  command_file_in($params, '"default.txt"');
  $settings->{output_text_fh}->print("# Created from $settings->{input_file} (", scalar(localtime), ")\r\n");
  $settings->{output_text_fh}->print($settings->{format_name}, "\r\n");

  read_branch($settings->{first_branch}, $params);

  delete $settings->{input_fh};
  $fh->close;
  delete $settings->{output_text_fh};
  delete $settings->{append_output_text_fh};
  delete $settings->{append_output_link_file};
  %database = ();
}

sub determine_format {
  my @branches = @_;
  my %formats;
  CHOICE: foreach my $choice (@branches) {
    if ($choice->{type} eq 'format') {
      die qq{Redefinition of format "$choice->{format}"!\n} if exists $formats{$choice->{format}};
      my $score = 0;
      my @params = @{$choice->{params}};
      while (defined (my $param = shift @params)) {

        if ($param eq 'extension') {
          my $extension = shift @params;
          if (lc substr($settings->{input_file}, -(length $extension)) ne lc $extension) {
            next CHOICE;
          }
        } elsif ($param eq 'string') {
          my ($offset, $text) = (shift(@params), string_unescape(shift(@params)));
          my $actual_text;
          my $old_pos = $settings->{input_fh}->tell;
          $settings->{input_fh}->seek($offset, 0);
          $settings->{input_fh}->read($actual_text, length($text));
          $settings->{input_fh}->seek($old_pos, 0);
          if ($actual_text ne $text) {
#            print "$actual_text ne $text\n";
            next CHOICE;
          }
        } else {
          die qq{Invalid format definition parameter "$param"!\n};
        }
        ++$score;
      }
      $formats{$choice->{format}}{score} = $score;

    } elsif ($choice->{type} eq 'branch') {
      foreach my $format (split /,\s*/, $choice->{params}[0]) {
        $formats{$format}{branch} = $choice;
      }
    } else {
      die "format_list must only contain format blocks and definitions!\n";
    }
  }
  my @scores = reverse sort map $formats{$_}{score}, grep exists $formats{$_}{score}, keys %formats;
  die qq{No possible format definitions detected!} if not @scores;
  my $max_scorer;
  foreach (grep exists $formats{$_}{score}, keys %formats) {
    if ($formats{$_}{score} == $scores[0]) {
      die qq{Cannot decide between multiple file formats!} if defined $max_scorer;
      $max_scorer = $_;
    }
  }
  die qq{No data for format $max_scorer!} unless exists $formats{$max_scorer}{branch};
  $settings->{format_name} = $max_scorer;
  return $formats{$max_scorer}{branch};
}

sub force_format {
  my ($required_format, @branches) = @_;
  foreach my $choice (@branches) {
    if ($choice->{type} eq 'branch') {
      foreach my $format (split /,\s*/, $choice->{params}[0]) {
        if ($format eq $required_format) {
          $settings->{first_branch} = $choice;
          $settings->{format_name} = $format;
          return;
        }
      }
    }
  }
  die qq{Unable to find forced format "$required_format"!\n};
}

sub read_branch {
  my $branch = shift;
  my $params = shift;

  if ($branch->{action} eq 'format') {
    print "Format $branch->{params}[0]\n" if DEBUG;
    read_branch_code($branch, $params);

  } elsif ($branch->{action} eq 'ignore') {
    print "Ignoring block\n" if DEBUG;

  } elsif ($branch->{action} eq 'repeat') {
    my $iterator = $branch->{params}[1];
    if ($iterator !~ /^[a-zA-Z0-9]$/) { die qq{Invalid repeat variable "$iterator" at line $branch->{line_no}!\n} }
    my $loop_count = to_value($branch->{params}[0], $params);
    print "Repeating block $loop_count times, iterating $iterator\n" if DEBUG;
    if ($params->{variables}{$iterator}) { die qq{Redefined repeat variable "$iterator" at line $branch->{line_no}!\n} }

    $params->{variables}{$iterator}{max} = $loop_count;
    $params->{variables}{$iterator}{value} = 0;
#    my $count = 0;
#    while ($count < $params->{variables}{$iterator}{max}) {
#      $params->{variables}{$iterator}{value} = $count++;
    while ($params->{variables}{$iterator}{value} < $params->{variables}{$iterator}{max}) {
      read_branch_code($branch, $params);
      ++$params->{variables}{$iterator}{value};
    }
    delete $params->{variables}{$iterator};

  } elsif ($branch->{action} eq 'local') {
    my $var = $branch->{params}[0];
    local $params->{variables}{$var} = undef;
    read_branch_code($branch, $params);

  } elsif ($branch->{action} eq 'if') {
    if ($branch->{params}[0] eq 'format') {
      if ($settings->{format_name} eq $branch->{params}[1]) {
        read_branch_code($branch, $params);
      }
    } elsif ($branch->{params}[0] eq 'reading') {
      read_branch_code($branch, $params);
    } elsif ($branch->{params}[0] eq 'writing') {
      # do nothing
    } else {
      if (to_value($branch->{params}[0], $params)) {
        read_branch_code($branch, $params);
      }
    }

  } else {
    die "Invalid branch action: $branch->{action}!\n";
  }
}

sub read_branch_code {
  my $branch = shift;
  my $params = shift;
  foreach my $line (@{$branch->{contents}}) {
    print "LINE $line->{line_no}: $line->{raw_line}\n" if SAY_LINE and $line->{raw_line};
    if ($line->{type} eq 'branch') {
      read_branch($line, $params);

    } elsif ($line->{type} eq 'command_in') {
      print "Running $line->{command} (", join(' ', @{$line->{params}}), ")\n" if HEAVY_DEBUG;
      $command_words{in}{$line->{command}}->($params, @{$line->{params}});

    } elsif ($line->{type} eq 'command_out') {
      # do nothing

    } elsif ($line->{type} eq 'read') {
      my ($field_name, $db_key) = interpolate_variables($line->{field}, $params);
      print "Reading $field_name (" . unpack('H*', $db_key). ") as $line->{data_type} (", join(' ', @{$line->{params}}), ")\n" if DEBUG;
      my $data = read_data($line->{data_type}, $line->{params}, $params);

      my @line_params = @{$line->{params}};
      for my $pid (0..$#line_params) {
        if ($line_params[$pid] eq 'compress') {
          my ($data_format) = @line_params[($pid+1)];
          $data = decompress($data, $data_format);
          splice @line_params, $pid, 2;
          last;
        }
      }

      db_store($db_key, $data) unless is_any('forget', @{$line->{params}});

      data_to_text($data, $field_name, $line->{data_type}, $line->{params}, $params) unless is_any('quiet', @{$line->{params}}) or is_any('hidden', @{$line->{params}});

      print "  Read ", unpack('H*', $data), "\n" if HEAVY_DEBUG;

    } else {
      die "Invalid line type: $line->{type}!";
    }
  }
}

sub string_escape {
  my ($text) = @_;
  $text =~ s/\\/\\\\/g; # \ -> \\
  $text =~ s/\0/\\0/g;  # [nul] -> \0
  $text =~ s/"/\\"/g;   # " -> \"
  $text =~ s/\r\n/\\n/g;# [crlf] -> \n
  $text =~ s/([\x00-\x1F\x7F-\xFF])/sprintf '\x%02x', ord $1/eg; # handle any remaining non-nice-ascii chars
  return qq{"$text"};
}

# This regexp version is buggy with input like '\\x00' etc :-(
#sub string_unescape { # reverse of string_escape
#  my ($text) = @_;
#  $text =~ s/^"(.*)"$/$1/g; # remove quotes
#  $text =~ s/\\x([0-9A-F]{2})/chr hex $1/eg; # \x__ -> any char
#  $text =~ s/\\n/\r\n/g;# \n -> [crlf]
#  $text =~ s/\\"/"/g;   # \" -> "
#  $text =~ s/\\0/\0/g;  # \0 -> [nul]
#  $text =~ s/\\\\/\\/g; # \\ -> \
#  return $text;
#}

sub string_unescape {
  my ($text) = @_;
  $text =~ s/^"(.*)"$/$1/g; # remove quotes
  my @chars = split //, $text;
  $text = '';
  my $escaped = 0;
  while (defined (my $char = shift @chars)) {
    if ($char eq '\\' and not $escaped) {
      $escaped = 1;
    } elsif ($escaped) {
      if ($char eq '\\' or $char eq '"') {
        $text .= $char;
      } elsif ($char eq '0') {
        $text .= "\0";
      } elsif ($char eq 'n') {
        $text .= "\r\n";
      } elsif ($char eq 'x') {
        $text .= chr(hex(shift(@chars).shift(@chars)));
      } else {
        die qq{Invalid escape sequence \\$char in input text\n  $text\n};
      }
      $escaped = 0;
    } else {
      $text .= $char;
    }
  }
  return $text;
}

sub data_to_text {
  my ($data, $field_name, $data_types, $line_params, $params) = @_;
  my @values;

  my $link_file = undef;
  my @line_params = @$line_params; # to splice safely without deleting the original data (I hope)
  for (0..$#line_params) {
    if ($line_params[$_] eq 'link') {
      $link_file = $line_params[$_+1];
      splice @line_params, $_, 2;
      last;
    }
  }

  my @to_be_escaped = ();

  foreach my $group (split(/,\s*/, $data_types)) {
    my ($type, $size) = (substr($group, 0, 1), substr($group, 1));
    if (not length $size) {
      $size = 1;
    } elsif (substr($size, 0, 1) eq '!') {
      $size = int(to_value(substr($size, 1), $params));
    } elsif ($type ne '@') {
      $size = int($size);
    }
    if ($type eq 'l' or $type eq 'L') {
      push @values, unpack $type => substr($data, 0, 4, '') for 1..$size;
    } elsif ($type eq 'f') {
      push @values, unpack_f(substr($data, 0, 4, '')) for 1..$size;
    } elsif ($type eq 'F') {
      push @values, unpack_f("\0\0".substr($data, 0, 2, '')) for 1..$size;
    } elsif ($type eq 's' or $type eq 'S') {
      push @values, unpack $type => substr($data, 0, 2, '') for 1..$size;
    } elsif ($type eq 'c' or $type eq 'C') {
      push @values, unpack $type => substr($data, 0, 1, '') for 1..$size;
    } elsif ($type eq 'B') {
      for (1..$size) {
        my $value = unpack 'B*' => substr($data, 0, 1, '');
        $value =~ tr/01/.!/;
        push @values, $value;
      }
    } elsif ($type eq 'x') {
      push @values, unpack 'H*' => substr($data, 0, $size, '');
    } elsif ($type eq 'z') {
      push @values, unpack 'Z*' => substr($data, 0, $size, '');
      push @to_be_escaped, $#values if not defined $link_file;
    } elsif ($type eq 'u') {
      push @values, pack 'C*' => unpack 'S*' => substr($data, 0, $size*2, '');
      push @to_be_escaped, $#values if not defined $link_file;
    } elsif ($type eq 'd') {
      push @values, substr($data, 0, $size, '');
      push @to_be_escaped, $#values if not defined $link_file;
    } elsif ($type eq 'a') {
      if (grep { $_ eq 'compress' } @line_params) { # allow compressed data to work properly (I hope)
        $size = length($data);
      }
      push @values, substr($data, 0, $size, '');
      push @to_be_escaped, $#values if not defined $link_file;
    } elsif ($type eq 'v') {
      return; # don't do anything at all
    } elsif ($type eq '@') {
      if ($size eq 'datetime') {
        push @values, join('/', unpack SCC => substr($data, 0, 4, '')).','.
                      join(':', map sprintf('%02d',$_), unpack CCC => substr($data, 0, 3, ''));
      } else {
        die qq{Can't textify invalid special data type "$size" in "$data_types"!\n};
      }
    } else {
      die qq{Can't textify invalid data type "$type" in "$data_types"!\n};
    }
  }

  for my $pos (0..$#line_params) {
    if ($line_params[$pos] eq 'image') {
      my ($image_format, $data_format, $width, $height) = @line_params[($pos+1)..($pos+4)];
      $_ = to_value($_, $params) foreach ($width, $height);
      $data_format = to_value($data_format, $params) if $data_format =~ /[\{\}]/;
      $_ = data_to_image($_, $image_format, $data_format, $width, $height) foreach @values;
      splice @line_params, $pos, 5;
      last;
    } elsif ($line_params[$pos] eq 'filter') {
      my $data_format = $line_params[$pos+1];
      $data_format = to_value($data_format, $params) if $data_format =~ /[\{\}]/;
      $_ = data_filter_out($_, $data_format) foreach @values;
      splice @line_params, $pos, 2;
      last;
    }
  }

  $values[$_] = string_escape($values[$_]) foreach @to_be_escaped;

  if (not defined $link_file) {
    $settings->{output_text_fh}->print("$field_name: ", join(' ', @values), "\r\n");
  } else {
    my $filename = to_value($link_file, $params);
    print ">$filename\t" if DEBUG;
    my $fh;
    if ($params->{append_output_link_file}{$filename}++) {
      $fh = create_output_fh($filename, '>>');
    } else {
      $fh = create_output_fh($filename);
    }
    $fh->print(join(' ', @values));
    $fh->close;
  }
}

sub write_using {
  my $tree = shift;
  $settings->{parse_tree} = $tree;

  if (ALLOW_PATCH and $settings->{patch_file}) {
    my $fh = IO::File->new("< $settings->{patch_file}") or
                           die qq{Unable to open $settings->{patch_file} for input: $!};
    my %patches;
    while (defined (my $line = $fh->getline)) {
      $line =~ s/\r?\n$//;
      $line =~ /^([a-zA-Z0-9_\-]+): (.*)$/ or die qq{Invalid patch line "$line"!\n};
      $patches{$1} = $2;
    }
    $settings->{patches} = \%patches;
  }

  my $fh = IO::File->new("> $settings->{output_file}") or
                         die qq{Unable to open $settings->{output_file} for output: $!};
  binmode $fh;
  $fh->autoflush(1) if $autoflush;
  $settings->{output_fh} = $fh;

  my $params = { action => 'write' };
  command_file_out($params, '"default.txt"');

  my $format;
  do { $format = $settings->{input_text_fh}->getline } until $format !~ /^#/;
  $format =~ s/\r?\n$//; # because chomp won't chomp the binmode's \r

  force_format($format, @{$tree->{contents}});

  print "Writing with format $settings->{format_name}\n" if DEBUG;

  $settings->{write_memory} = {};

  my ($compression, $comp_data) = (undef, '');
  my @format_params = @{$settings->{first_branch}{params}};
  for my $pid (0..$#format_params) {
    if (defined $format_params[$pid] and $format_params[$pid] eq 'compression') {
      $compression = $format_params[$pid+1];
      $settings->{output_fh} = new IO::Scalar \$comp_data;
      $settings->{output_scalar_ref} = \$comp_data;

      splice @format_params, $pid, 2;
      redo;
    }
  }

  write_branch($settings->{first_branch}, $params);

  if ($compression) {
    $fh->print(compress($comp_data, $compression));
  }

  delete $settings->{write_memory};
  delete $settings->{output_fh};
  $fh->close;
  delete $settings->{input_text_fh};
  delete $settings->{append_input_text_fh};
  %database = ();
}

sub write_branch {
  my $branch = shift;
  my $params = shift;

  print "LINE $branch->{line_no}: $branch->{raw_line}\n" if SAY_LINE and $branch->{raw_line};

  if ($branch->{action} eq 'format') {
    print "Format $branch->{params}[0]\n" if DEBUG;
    write_branch_code($branch, $params);

  } elsif ($branch->{action} eq 'ignore') {
    print "Ignoring block\n" if DEBUG;

  } elsif ($branch->{action} eq 'repeat') {
    my $iterator = $branch->{params}[1];
    my $loop_count = to_value($branch->{params}[0], $params);
    print "Repeating block $loop_count times, iterating $iterator\n" if DEBUG;
    if ($iterator !~ /^[a-zA-Z0-9]$/) { die qq{Invalid repeat variable "$iterator" at line $branch->{line_no}!\n} }
    if ($params->{variables}{$iterator}) { die qq{Redefined repeat variable "$iterator" at line $branch->{line_no}!\n} }

    $params->{variables}{$iterator}{max} = $loop_count;
    my $count = 0;
    while ($count < $params->{variables}{$iterator}{max}) {
      $params->{variables}{$iterator}{value} = $count++;
      write_branch_code($branch, $params);
    }
    delete $params->{variables}{$iterator};

  } elsif ($branch->{action} eq 'local') {
    my $var = $branch->{params}[0];
    local $params->{variables}{$var} = undef;
    write_branch_code($branch, $params);

  } elsif ($branch->{action} eq 'if') {
    if ($branch->{params}[0] eq 'format') {
      if ($settings->{format_name} eq $branch->{params}[1]) {
        write_branch_code($branch, $params);
      }
    } elsif ($branch->{params}[0] eq 'reading') {
      # do nothing
    } elsif ($branch->{params}[0] eq 'writing') {
      write_branch_code($branch, $params);
    } else {
      if (to_value($branch->{params}[0], $params)) {
        write_branch_code($branch, $params);
      }
    }

  } else {
    die "Invalid branch action: $branch->{action}!\n";
  }
}

sub write_branch_code {
  my $branch = shift;
  my $params = shift;
  foreach my $line (@{$branch->{contents}}) {
    print "LINE $line->{line_no}: $line->{raw_line}\n" if SAY_LINE and $line->{raw_line};
    if ($line->{type} eq 'branch') {
      write_branch($line, $params);

    } elsif ($line->{type} eq 'command_in') {
      # do nothing

    } elsif ($line->{type} eq 'command_out') {
      print "Running $line->{command} (", join(' ', @{$line->{params}}), ")\n" if HEAVY_DEBUG;
      $command_words{out}{$line->{command}}->($params, @{$line->{params}});

    } elsif ($line->{type} eq 'read') {
      my ($field_name, $db_key) = interpolate_variables($line->{field}, $params);
      print "Writing $field_name (" . unpack('H*', $db_key). ") as $line->{data_type} (", join(' ', @{$line->{params}}), ")\n" if HEAVY_DEBUG;

      if (grep $_ eq '?', @{$line->{params}}) {
        #die qq{Memorised fields must be a single value (not $line->{data_type})! (line $line->{line_no})} if index($line->{data_type}, ',') != -1;
        $settings->{write_memory}{$db_key} = [ $settings->{output_fh}->tell, $line->{data_type} ];
      }
      if (is_any('hidden', @{$line->{params}})) {
        my $data = invent_text_data($field_name, $line->{data_type}, $line->{params}, $params);
        print "  Wrote ", unpack('H*', $data), "\n" if HEAVY_DEBUG;
        db_store($db_key, $data);
        $settings->{output_fh}->print($data) unless $line->{data_type} =~ /^v/;
      } elsif (not is_any('quiet', @{$line->{params}})) {
        my $data = read_text_data($field_name, $line->{data_type}, $line->{params}, $params);
        print "  Wrote ", unpack('H*', $data), "\n" if HEAVY_DEBUG;
        db_store($db_key, $data);
        $settings->{output_fh}->print($data) unless $line->{data_type} =~ /^v/;
      }
    } else {
      die "Invalid line type: $line->{type}!";
    }
  }
}

sub invent_text_data {
  my ($field_name, $data_types, $line_params, $params) = @_;
  my @data_types = split(/,\s*/, $data_types);
  die "Can't invent !=1 data item!" if @data_types != 1;
  my @values;
  if ($data_types[0] =~ /^[adzu]/) {
    push @values, '';
  } else {
    push @values, '0';
  }
  return text_to_data(\@data_types, \@values, $line_params, $params);
}

sub read_text_data {
  my ($field_name, $data_types, $line_params, $params) = @_;
  my @data_types = split(/,\s*/, $data_types);
  my @values;

  my $link_file = undef;
  for (0..$#$line_params) {
    if ($line_params->[$_] eq 'link') {
      $link_file = $line_params->[$_+1];
      last;
    }
  }

  if ($data_types =~ /^v/) {
    # void/variable - don't read any data
    print qq{[not] Ignoring "$data_types"\n} if HEAVY_DEBUG;
    return to_value(substr($data_types, 2), $params);

  } elsif (defined $link_file) {
    my $filename = to_value($link_file, $params);
    print "<$filename\t" if DEBUG;
    my $fh = create_input_fh($filename);
    @values = (join '', $fh->getlines);
    $fh->close;

  } else {
    my $line;
    do { $line = $settings->{input_text_fh}->getline } until $line !~ /^#/;
    $line =~ s/\r?\n$//; # because chomp won't chomp the \r from binary mode
    $line =~ /^$field_name: (.*)$/ or die qq{Invalid input line (does not start with "$field_name: "):\n  "$line"\n};
    my $line_data = $1;
    if (ALLOW_PATCH and $settings->{patches} and exists $settings->{patches}{$field_name}) {
      print "Patching $line_data --> $settings->{patches}{$field_name}\n" if HEAVY_DEBUG;
      $line_data = $settings->{patches}{$field_name};
    }
    @values = ('');
    my $in_string = 0;
    my $escaped = 0;
    my $done_anything = 0;
    my @chars = split //, $line_data;
    while (defined (my $char = shift @chars)) {
      $done_anything = 1;
      if ($char eq ' ' and not $in_string) {
        push @values, '';
      } elsif ($char eq '"' and not $in_string) {
        $in_string = 1;
      } elsif ($char eq '"' and $in_string and not $escaped) {
        $in_string = 0;
      } elsif ($char eq '\\' and not $escaped) {
        $escaped = 1;
      } elsif ($escaped) {
        if ($char eq '\\' or $char eq '"') {
          $values[-1] .= $char;
        } elsif ($char eq '0') {
          $values[-1] .= "\0";
        } elsif ($char eq 'n') {
          $values[-1] .= "\r\n";
        } elsif ($char eq 'x') {
          $values[-1] .= chr(hex(shift(@chars).shift(@chars)));
        } else {
          die qq{Invalid escape sequence \\$char in input line\n  "$line"\n};
        }
        $escaped = 0;
      } else {
        $values[-1] .= $char;
      }
    }
    pop @values if not $done_anything; # remove an empty final value
  }

  my @line_params = @$line_params;
  for (0..$#line_params) {
    if ($line_params[$_] eq 'compress') {
      my ($data_format) = @line_params[($_+1)];
      $_ = compress($_, $data_format) foreach @values;
      splice @line_params, $_, 2;
      last;
    }
  }
  for (0..$#line_params) {
    if ($line_params[$_] eq 'image') {
      my ($image_format, $data_format, $width, $height) = @line_params[($_+1)..($_+4)];
#      $_ = to_value($_, $params) foreach ($width, $height); # (is there any point in these when converting from the image file to raw data?)
      $_ = image_to_data($_, $image_format, $data_format) foreach @values;
      splice @line_params, $_, 5;
      last;
    } elsif ($line_params[$_] eq 'filter') {
      my $data_format = $line_params[$_+1];
      $data_format = to_value($data_format, $params) if $data_format =~ /[\{\}]/;
      $_ = data_filter_in($_, $data_format) foreach @values;
      splice @line_params, $_, 2;
      last;
    }
  }

  return text_to_data(\@data_types, \@values, $line_params, $params);
}

sub text_to_data {
  my ($data_types, $values, $line_params, $params) = @_;
  my $return = '';

#  return $return if @$values == 0;

  foreach my $group (@$data_types) {
    my ($type, $size) = (substr($group, 0, 1), substr($group, 1));
    if (not length $size) {
      $size = 1;
    } elsif (substr($size, 0, 1) eq '!') {
      $size = int(to_value(substr($size, 1), $params));
    } elsif ($type ne '@') {
      $size = int($size);
    }

    if ($type eq 'l' or $type eq 'L' or
        $type eq 's' or $type eq 'S' or
        $type eq 'c' or $type eq 'C') {
      $return .= pack $type => shift @$values for 1..$size;
    } elsif ($type eq 'f') {
      $return .= pack_f(shift @$values) for 1..$size;
    } elsif ($type eq 'F') {
      $return .= substr pack_f(shift @$values), 2 for 1..$size;
    } elsif ($type eq 'B') {
      $return .= pack 'B*' => shift @$values for 1..$size;
    } elsif ($type eq 'x') {
      $return .= pack 'H*' => shift @$values;
    } elsif ($type eq 'z') {
      $return .= pack 'Z*' => shift @$values;
    } elsif ($type eq 'u') {
      $return .= pack 'S*' => unpack 'C*' => shift @$values;
    } elsif ($type eq 'd') {
      $return .= pack "a$size" => shift @$values;
    } elsif ($type eq 'a') {
      $return .= pack "a*" => shift @$values;
    } elsif ($type eq 'v') {
      return $size;
      # do nothing
    } elsif ($type eq '@') {
      if ($size eq 'datetime') {
        $return .= pack SCCCCC => split m~[/:,]~, shift @$values;
      } else {
        die qq{Can't understand special data type "$size" in "$data_types"!\n};
      }
    } else {
      die qq{Can't understand invalid data type "$type" in "$data_types"!\n};
    }
  }
  die qq{Input values (@$values) don't match required format (@$data_types)!\n} if @$values;
  return $return;
}

sub db_store { $database{$_[0]} = $_[1] }
sub db_read  { $database{$_[0]}         }
sub db_delete{  undef $database{$_[0]};
               delete $database{$_[0]}  }

sub parse_format {
  my @format = @_;

  %fields = (); # reset it

  my ($tree, $branch, @parsed);
  my $line_no = 0;
  foreach my $line (@format) {
    ++$line_no;
    chomp $line;
    my $raw_line = $line;
    $line =~ s/#.*$//; # strip comments
    $line =~ s/^\s*(.*?)\s*$/$1/; # strip surrounding whitespace
    next unless length $line; # ignore empty lines
    my @words = split /\s+/, $line;
    UNSPLIT: {
      for my $id (0..$#words) {
        if (substr($words[$id], -1) eq '{') { # rejoin code blocks
          my $closing_brace = -1;
          for my $id2 ($id+1..$#words) {
            if ($words[$id2] eq '}') {
              $closing_brace = $id2;
              last;
            }
          }
          die qq{Cannot find closing brace in line $line_no:\n  "$raw_line"\n} if $closing_brace == -1;
          splice @words, $id, 1+$closing_brace-$id, join ' ', @words[$id..$closing_brace];
          redo UNSPLIT;
        }
        if ($words[$id] =~ /,$/) { # rejoin comma-separated words
          splice @words, $id, 2, join(' ', @words[$id,$id+1]);
          redo UNSPLIT;
        }
      }
    }
    if ($words[0] eq '[') { # grouped statements
      my $new_branch = { type => 'branch', action => $words[1], params => [@words[2..$#words]], contents => [], parent => $branch, line_no => $line_no, raw_line => $raw_line };
      push @{$branch->{contents}}, $new_branch;
      $branch = $new_branch;
      $tree ||= $branch;
      $settings->{format_name} = (split /,/,$words[2])[0] if $words[1] eq 'format';
    } elsif ($words[0] eq ']') {
      die qq{Unmatched ] at line $line_no!\n} unless $branch;
      $branch = $branch->{parent};
    } else {
      die qq{Line $line_no isn't inside any tree!\n} unless $branch;

      my $is_command = 0;

      if ($words[0] eq '<' or $words[0] eq '*') { # input command
        $is_command = 1;
        my ($command, @params) = @words[1..$#words];
        die qq{Invalid input command word "$command" in line $line_no:\n  "$raw_line"\n} unless $command_words{in}{$command};
        push @{$branch->{contents}}, { type => 'command_in', command => $command, params => \@params, line_no => $line_no, raw_line => $raw_line };
      }
      if ($words[0] eq '>' or $words[0] eq '*') { # output command
        $is_command = 1;
        my ($command, @params) = @words[1..$#words];
        die qq{Invalid output command word "$command" in line $line_no:\n  "$raw_line"\n} unless $command_words{out}{$command};
        push @{$branch->{contents}}, { type => 'command_out', command => $command, params => \@params, line_no => $line_no, raw_line => $raw_line };
      }
      if (not $is_command) { # normal read
        if ($words[0] eq '~') { # format definition
          if ($words[1] eq 'alias') {
            $settings->{format_alias}{$_} = $words[2] foreach @words[3..$#words];
          } else {
            push @{$branch->{contents}}, { type => 'format', format => $words[1], params => [@words[2..$#words]], line_no => $line_no, raw_line => $raw_line };
          }
        } else {
          push @{$branch->{contents}}, { type => 'read', field => add_field($words[0], $words[1]), data_type => $words[1], params => [@words[2..$#words]], line_no => $line_no, raw_line => $raw_line };
        }
      }
    }
  }
  die qq{Unmatched [ somewhere!\n} if $branch;
  return $tree;
}

=pod
  sub deparse_branch {
    my $branch = shift;
    my $indent = shift || '';
    print "$indent\[ $branch->{action}\n";
    foreach my $line (@{$branch->{contents}}) {
      if ($line->{type} eq 'branch') {
        deparse_branch($line, "$indent  ");
      } elsif ($line->{type} =~ /command/) {
        print "$indent  * $line->{command} ", join(' ', @{$line->{params}}), "\n";
      } elsif ($line->{type} eq 'format') {
        print "$indent  format $line->{format} ", join(' ', @{$line->{params}}), "\n";
      } elsif ($line->{type} eq 'read') {
        print "$indent  $fields{from_id}[$line->{field}] $line->{data_type} ", join(' ', @{$line->{params}}), "\n";
      } else {
        die qq{Invalid branch type "$line->{type}"!};
      }
    }
    print "$indent]\n";
  }
=cut

sub add_field {
  my $field = shift;
  my $type = shift;

  if (exists $fields{to_id}{$field}) {
    if ($fields{type}[$fields{to_id}{$field}] ne (split(/,/, $type))[0]) {
      $fields{type_warnings}[$fields{to_id}{$field}] = 1;
    }
    return $fields{to_id}{$field}; # allow redefinitions
  }

  push @{$fields{from_id}}, $field;
  push @{$fields{type}}, (split(/,/, $type))[0];
  $fields{to_id}{$field} = $#{$fields{from_id}};
  return $fields{to_id}{$field};
}

sub to_value {
  my ($expression, $params) = @_;
  print qq{To_val'ing $expression\n} if HEAVY_DEBUG;
  if ($expression =~ /^[0-9]+$/) {
    return $expression;
  } elsif ($expression =~ /^"(.*)"$/) {
    return $1;
  } elsif ($expression =~ /^\{(.*)\}$/) {
    return safe_eval($1, $params);
  } else {
    if (exists $fields{to_id}{$expression}) {
      return get_field_value($fields{to_id}{$expression}, $params);
    } else {
      die qq{Cannot find variable "$expression"!\n};
    }
  }
}

sub get_field_value {
  my ($id, $params) = @_;
  if ($fields{type_warnings}[$id]) {
    die "Field $fields{from_id}[$id] has multiple defined types!";
  }
  my $field_type = $fields{type}[$id];
  my ($text, $full_key) = interpolate_variables($id, $params);

  my $db_data = db_read($full_key);
  return undef unless defined $db_data;

  my ($type, $size) = (substr($field_type, 0, 1), substr($field_type, 1));
  if (not length $size) {
    $size = 1;
  } elsif (substr($size, 0, 1) eq '!') {
    $size = int(to_value(substr($size, 1), $params));
  } elsif ($type ne '@') {
    $size = int($size);
  }

  if (is_any($type, qw(l L s S c C))) {
    return unpack substr($field_type, 0, 1), $db_data;

  } elsif ($type eq 'f') {
    return unpack_f($db_data);

  } elsif ($type eq 'F') {
    return unpack_f("\0\0".$db_data);

  } elsif ($type eq 'z') {
    return unpack "Z*", $db_data;

  } elsif ($type eq 'u') {
    return pack 'C*', unpack 'S*', $db_data;

  } elsif ($type eq 'd') {
    return $db_data;

  } elsif ($type eq 'a') {
    return $db_data;

  } elsif ($type eq 'x') {
    return unpack "h$size", $db_data;

  } elsif ($type eq 'v') {
    print "Returning (v) $db_data\n" if HEAVY_DEBUG;
    return $db_data;

  } elsif ($type eq '@') {
    my $special_type = substr($field_type, 1);
    if ($special_type eq 'datetime') {
      my $date = $db_data;
      return join('/', unpack SCC => substr($date, 0, 4, '')).','.
             join(':', map sprintf('%02d',$_), unpack CCC => substr($date, 0, 3, ''));
    } else {
      die qq{Invalid special data type "$special_type" in "$text"!\n};
    }

  } else {
    die qq{Invalid type "$field_type" of $text!\n};
  }
}

sub is_any {
  my $text = shift;
  foreach (@_) { return 1 if $text eq $_ }
  return 0;
}

sub interpolate_variables {
  my ($id, $params) = @_;
#  confess unless defined $id;
  my @possible_names = ($fields{from_id}[$id]);
  push @possible_names, @{$aliases{$possible_names[0]}} if $aliases{$possible_names[0]};
  print qq{Checking reverse @possible_names\n} if HEAVY_DEBUG;
  print qq{with variables: }, join(',',keys %{$params->{variables}}), qq{\n} if HEAVY_DEBUG;
  foreach my $text (reverse @possible_names) { # use reverse so the latest defined alias is used where possible
    my $full_key = pack FIELD_PACK_FORMAT, $id;
    my $index;
    foreach my $var (sort keys %{$params->{variables}}) {
      if (($index = index($text, "%$var")) != -1) {
        substr($text, $index, 2) = $params->{variables}{$var}{value};
        $full_key .= pack FIELD_VAR_PACK_FORMAT, $params->{variables}{$var}{value};
      }
    }
    print qq{Field $id matching $text - } if HEAVY_DEBUG;
    if (index($text, "%") == -1) {
      print qq{succeeded ($full_key)\n} if HEAVY_DEBUG;
      return ($text, $full_key);
    }
    print qq{failed\n} if HEAVY_DEBUG;
  }
  die qq{"$fields{from_id}[$id]" called without all variables defined!\n};
}

sub safe_eval {
  my ($code, $params) = @_;
  my @new_code;
  foreach my $word (split /\s+/, $code) {
    if ($word eq '<here>') {
      if ($params->{action} eq 'read') {
        push @new_code, $settings->{input_fh}->tell;
      } else {
        push @new_code, $settings->{output_fh}->tell;
      }
    } elsif ($word eq '<filelen>') {
      if ($params->{action} eq 'read') {
        push @new_code, $settings->{input_fh_length};
      } else {
        push @new_code, -1;
      }
    } elsif ($word =~ /^<(\S+)>$/) {
      push @new_code, get_field_value($fields{to_id}{$1}, $params);
    } elsif ($word =~ /^length\(<(.*)>\)$/) {
      push @new_code, length(get_field_value($fields{to_id}{$1}, $params));
    } elsif ($word =~ /^'<(\S+)>'$/) {
      my $new_word = get_field_value($fields{to_id}{$1}, $params);
      $new_word = '' if not defined $new_word;
      $new_word =~ s/([\x00-\x1F\x7F-\xFF"\$\@\\])/sprintf '\x%02x', ord $1/eg; # convert unpleasant characters to \xFF etc
      push @new_code, qq{"$new_word"};
    } elsif ($word =~ /^%([a-zA-Z0-9])$/) {
      push @new_code, $params->{variables}{$1}{value};
    } else {
      push @new_code, $word;
    }
  }
  $code = join ' ', @new_code;
  print qq{Evalling "$code"\n} if HEAVY_DEBUG;
  if ($code =~ /^\s*([0-9\.]+)\s*$/) { # don't bother evalling simple numbers
    print qq{(quick return)\n} if HEAVY_DEBUG;
    return $1;
  } else {
    my $safe = new Safe;
    $safe->permit_only(qw(:base_core :base_mem :base_loop :base_orig :base_math));
    $safe->share(qw(%advanced_flags));
    my $return = $safe->reval($code);
    die qq{Error executing code "$code":\n  $@\n} if $@;
    return $return;
  }
}

sub input_fh_read {
  my $length = shift;
  my $params = shift;
  my $return;
  my $read = $settings->{input_fh}->read($return, $length);
  if ($read != $length and not $params->{ignore_eof}) {
    carp "Warning: probably read past EOF!\n" if $EOF_count++ < 10;
    print "[ignoring EOFs now]\n" if $EOF_count == 10;
  }
  return $return;
}

sub read_data {
  my ($data_types, $line_params, $params) = @_;
  my $return;
  my $read_params = {};
  $read_params->{ignore_eof} = 1 if grep @$line_params, 'ignore_eof';
  foreach my $group (split(/,\s*/, $data_types)) {
    my ($type, $size) = (substr($group, 0, 1), substr($group, 1));
    if (not length $size) {
      $size = 1;
    } elsif (substr($size, 0, 1) eq '!') {
      $size = int(to_value(substr($size, 1), $params));
    } elsif ($type ne '@') {
      $size = int($size);
    }
    if ($type eq 'l' or $type eq 'L' or $type eq 'f') {
      $return .= input_fh_read(4, $read_params) for 1..$size;
    } elsif ($type eq 's' or $type eq 'S' or $type eq 'F') {
      $return .= input_fh_read(2, $read_params) for 1..$size;
    } elsif ($type eq 'c' or $type eq 'C' or $type eq 'B') {
      $return .= input_fh_read(1, $read_params) for 1..$size;
    } elsif ($type eq 'x' or $type eq 'z' or $type eq 'd' or $type eq 'a') {
      $return .= input_fh_read($size, $read_params);
    } elsif ($type eq 'u') {
      $return .= input_fh_read($size*2, $read_params);
    } elsif ($type eq 'v') {
      $return .= $size;
    } elsif ($type eq '@') {
      if ($size eq 'datetime') {
        $return .= input_fh_read(7, $read_params);
      } else {
        die qq{Can't read invalid special data type "$size" in "$data_types"!\n};
      }
    } else {
      die qq{Can't read invalid data type "$type" in "$data_types"!\n};
    }
  }
  return $return;
}

sub command_seek {
  my $params = shift;
  my ($pos) = @_;
  $pos = to_value($pos, $params);
  $settings->{input_fh}->seek($pos, 0);
}

sub command_write {
  my $params = shift;
  my ($field, $value) = @_;

  $value = to_value($value, $params);
  my ($field_name, $db_key) = interpolate_variables($fields{to_id}{$field}, $params);
  my ($field_pos, $data_type) = @{$settings->{write_memory}{$db_key}};

  my $data = text_to_data([$data_type], [$value], {}, $params);
  print qq{Store "$data" into "$db_key"\n} if HEAVY_DEBUG;
  db_store($db_key, $data);
  if (not $data_type =~ /^v/) {
    if ($settings->{output_scalar_ref}) {
      # since IO::Scalar's broken if you try to overwrite data
      substr(${$settings->{output_scalar_ref}}, $field_pos, length($data)) = $data;
    } else {
      my $old_pos = $settings->{output_fh}->tell;
      $settings->{output_fh}->seek($field_pos, 0);
      $settings->{output_fh}->print($data);
      $settings->{output_fh}->seek($old_pos, 0);
     }
  }
}

sub command_file_in {
  my $params = shift;
  my ($filename) = @_;
  $filename = to_value($filename, $params);
  print "Redirecting output to $filename\n" if DEBUG;

  if (grep ref($_), values %{$params->{append_output_text_fh}} > FH_CACHE_LIMIT) { # limit filehandle cache to 10 items
    foreach my $key (grep ref($_), keys %{$params->{append_output_text_fh}}) {
      $params->{append_output_text_fh}{$key}->close;
      $params->{append_output_text_fh}{$key} = 0;
    }
  }

  if (exists $params->{append_output_text_fh}{$filename}) {
    if (ref $params->{append_output_text_fh}{$filename}) {
      $settings->{output_text_fh} = $params->{append_output_text_fh}{$filename};
    } else {
      my $fh = create_output_fh($filename, '>>');
      $params->{append_output_text_fh}{$filename} = $settings->{output_text_fh} = $fh;
    }
  } else {
    my $fh = create_output_fh($filename);
    $params->{append_output_text_fh}{$filename} = $settings->{output_text_fh} = $fh;
  }
}

sub command_file_out {
  my $params = shift;
  my ($filename) = @_;
  $filename = to_value($filename, $params);
  print "Redirecting input to $filename\n" if DEBUG;

  if (grep ref($_), values %{$params->{append_input_text_fh}} > FH_CACHE_LIMIT) { # limit filehandle cache to 10 items
    foreach my $key (grep ref($_), keys %{$params->{append_input_text_fh}}) {
      my $pos = $params->{append_input_text_fh}{$key}->tell;
      $params->{append_input_text_fh}{$key}->close;
      $params->{append_input_text_fh}{$key} = $pos;
    }
  }

  if (exists $params->{append_input_text_fh}{$filename}) {
    if (ref $params->{append_input_text_fh}{$filename}) {
      $settings->{input_text_fh} = $params->{append_input_text_fh}{$filename};
    } else {
      my $fh = create_input_fh($filename);
      $fh->seek($params->{append_output_text_fh}{$filename}, 0);
      $params->{append_input_text_fh}{$filename} = $settings->{input_text_fh} = $fh;
    }
  } else {
    my $fh = create_input_fh($filename);
    $params->{append_input_text_fh}{$filename} = $settings->{input_text_fh} = $fh;
  }
}

sub command_filetime_in {
  my $params = shift;
  my ($filename, $date) = @_;
  $filename = to_value($filename, $params);
  $date = to_value($date, $params);
  $date =~ m{^(\d+)/(\d+)/(\d+),(\d+):(\d+):(\d+)$} or die qq{"$date" is not a valid datetime value!\n};
  $date = timelocal($6, $5, $4, $3, $2-1, $1-1900);
  my $filepath = File::Spec->canonpath($settings->{output_text_path}.'\\'.$filename);
  utime $date, $date, $filepath;
}

sub command_filetime_out {
  my $params = shift;
  my ($filename, $date_field) = @_;
  $filename = to_value($filename, $params);
  my $filepath = File::Spec->canonpath($settings->{input_text_path}.'\\'.$filename);
  my @date = localtime((lstat $filepath)[9]);
  command_write($params, $date_field, '"'.join('/', $date[5]+1900, $date[4]+1, $date[3]).','.join(':', reverse @date[0..2]).'"');
}

sub command_repeat_alter {
  my $params = shift;
  my ($variable, $amount) = @_;
  $params->{variables}{$variable}{max} += to_value($amount, $params);
}

sub command_var_set {
  my $params = shift;
  my ($variable, $value) = @_;
  $params->{variables}{$variable}{value} = to_value($value, $params);
}
sub command_var_undef {
  my $params = shift;
  my ($variable, $value) = @_;
  delete $params->{variables}{$variable};
}

sub command_alias {
  my $params = shift;
  my ($new_name, $old_name) = @_;
  print qq{Aliasing $new_name to $old_name ($fields{to_id}{$old_name})\n} if HEAVY_DEBUG;
  if (not $aliases{$old_name} or not is_any($new_name, @{$aliases{$old_name}})) {
    $fields{to_id}{$new_name} = $fields{to_id}{$old_name};
    push @{$aliases{$old_name}}, $new_name;
    print qq{now @{$aliases{$old_name}}\n} if HEAVY_DEBUG;
  }
}

sub command_die {
 my $params = shift;
 my ($message) = @_;
 $message = to_value($message, $params);
 die qq{Format died: "$message"\n};
}

sub command_echo {
 my $params = shift;
 my ($message) = @_;
 $message = to_value($message, $params);
 log_add($message);
}

sub command_note_in {
 my $params = shift;
 my ($message) = @_;
 $message = to_value($message, $params);
 $settings->{output_text_fh}->print("# $message\r\n");
}

sub command_subformat_in {
  my $params = shift;
  my ($subformat, $data) = @_;
  $data = to_value($data, $params);
  $subformat = to_value($subformat, $params) unless $subformat =~ /^[a-z_0-9]+$/;
  $subformat = $settings->{format_alias}{$subformat} if defined $settings->{format_alias}{$subformat};
  print "Switching to format $subformat\n" if DEBUG;

  local $settings->{input_fh} = new IO::Scalar \$data;
  local $settings->{input_fh_length} = length($data);
  force_format($subformat, @{$settings->{parse_tree}{contents}});
  read_branch($settings->{first_branch}, $params);
  my $dif = $settings->{input_fh}->tell - length($data);
  if ($dif > 0) {
    print "Warning: subformat $subformat read $dif bytes too much!\n";
  } elsif ($dif < 0) {
    $dif = abs $dif;
    print "Warning: subformat $subformat read $dif bytes too little!\n";
  }
}
sub command_subformat_out {
  my $params = shift;
  my ($subformat, $data) = @_;
  $subformat = to_value($subformat, $params) unless $subformat =~ /^[a-z_0-9]+$/;
  $subformat = $settings->{format_alias}{$subformat} if defined $settings->{format_alias}{$subformat};
  print "Switching to format $subformat\n" if DEBUG;

  force_format($subformat, @{$settings->{parse_tree}{contents}});
  write_branch($settings->{first_branch}, $params);
}

sub command_forget {
  my $params = shift;
  my ($field) = @_;
  my ($field_name, $db_key) = interpolate_variables($fields{to_id}{$field}, $params);
  db_delete($db_key);
}

sub command_ensure {
  my $params = shift;
  my ($expr) = @_;
  my $value = to_value($expr, $params);
  die qq{Failed test "$expr"!\n} unless $value;
}

sub command_set_palette {
  my $params = shift;
  my ($format, $data) = @_;
  $data = to_value($data, $params);
  my @palette;
  if ($format eq 'rgbx') {
    @palette = map substr($data, 4*$_, 3), 0..length($data)/4-1;
  } elsif ($format eq '16') {
    @palette = map colour_16_to_bytes($_), unpack 'S*', $data;
  } elsif ($format eq '15') {
    @palette = map colour_15_to_bytes($_), unpack 'S*', $data;
  } elsif ($format eq '12') {
    @palette = map colour_12_to_bytes($_), unpack 'S*', $data;
  } else {
    die qq{Invalid palette format "$format"!};
  }
  $settings->{_image_palette} = \@palette;
}


sub create_output_fh {
  my $filename = shift;
  my $mode = shift || '>';
  my $filepath = File::Spec->canonpath($settings->{output_text_path}.'\\'.$filename);
  my @paths = File::Spec->splitdir($filepath);
  for (0..$#paths-1) {
    mkdir join('/', @paths[0..$_]) or warn "Error creating ".join('/', @paths[0..$_]).": $!" unless -e join('/', @paths[0..$_]);
  }
  log_add("Writing $filepath") if $mode eq '>';
  my $fh = IO::File->new("$mode $filepath") or
                         die qq{Unable to open $filepath for output: $!};
  binmode $fh;
  $fh->autoflush(1) if $autoflush;
  return $fh;
}

sub create_input_fh {
  my $filename = shift;
  my $filepath = File::Spec->canonpath($settings->{input_text_path}.'\\'.$filename);
  log_add("Reading $filepath");
  my $fh = IO::File->new("< $filepath") or
                         die qq{Unable to open $filepath for input: $!};
  binmode $fh;
  return $fh;
}

sub compress {
  my ($data, $data_format) = @_;
  if ($data_format eq 'zlib') {
    return compress__zlib_max($data);
  } elsif ($data_format eq 'l33t_zlib') {
    return 'l33t' . pack(L => length $data) . compress__zlib($data, Compress::Zlib::Z_DEFAULT_COMPRESSION()); # or Z_BEST_COMPRESSION
  } else {
    die qq{Invalid data compression format "$data_format"!\n};
  }
}
sub compress__zlib {
  my $data = $_[0];
  my $compression_level = $_[1];
  my ($deflate, $out1, $out2, $status);
  ($deflate, $status) = Compress::Zlib::deflateInit(-Level => $compression_level);
  die "deflateInit failed ($status)" if $status != Compress::Zlib::Z_OK();
  ($out1, $status) = $deflate->deflate(\$data);
  die "deflate failed ($status)" if $status != Compress::Zlib::Z_OK();
  ($out2, $status) = $deflate->flush;
  die "flush failed ($status)" if $status != Compress::Zlib::Z_OK();
  return $out1.$out2;
}
sub decompress {
  my ($data, $data_format) = @_;

  if ($data_format eq 'zlib') {
    my ($inflate, $out, $status);
    ($inflate, $status) = Compress::Zlib::inflateInit;
    die "inflateInit failed ($status)" if $status != Compress::Zlib::Z_OK();
    ($out, $status) = $inflate->inflate(\$data);
    die "inflate failed ($status)" if $status != Compress::Zlib::Z_STREAM_END();
    return $out;

  } elsif ($data_format eq 'l33t_zlib') {
    my $new_data = substr($data, 8); # 'l33t' followed by uncompressed size
    my ($inflate, $out, $status);
    ($inflate, $status) = Compress::Zlib::inflateInit;
    die "inflateInit failed ($status)" if $status != Compress::Zlib::Z_OK();
    ($out, $status) = $inflate->inflate(\$new_data);
    die "inflate failed ($status)" if $status != Compress::Zlib::Z_STREAM_END();
    return $out;

  } elsif ($data_format =~ /^bzone_(\d+)$/) {
    return bzone_decompress($data, $1);

  } else {
    die qq{Invalid data decompression format "$data_format"!\n};
  }
}

sub data_to_image {
  my ($data, $image_format, $data_format, $width, $height) = @_;
  my $raw_data = '';

  # first convert to a simple RGB format

  if ($data_format eq 'RGB0') {
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, ($y*$width + $x)*4, 3);
      }
    }

  } elsif ($data_format eq 'grey') {
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, $y*$width + $x, 1) x 3;
      }
    }

  } elsif ($data_format eq 'paletted') {
    my @palette = @{$settings->{_image_palette}} or die "Can't read paletted image with no palette!";
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= $palette[unpack C => substr($data, $y*$width + $x, 1)];
      }
    }

  } elsif ($data_format =~ /^grey_x(\d+)$/) {
    my $mult = $1;
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= (pack C => $mult * unpack C => substr($data, $y*$width + $x, 1)) x 3;
      }
    }

  } elsif ($data_format eq 'BDT') { # pairs of bytes (e.g. colour and bumpmap), split to produce two greyscale images
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, ($y*$width + $x)*2, 1) x 3;
      }
    }
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, ($y*$width + $x)*2+1, 1) x 3;
      }
    }
    $height *= 2;

  } elsif ($data_format =~ /^DDT_(\d)(3)$/) {
    my @pixel_format = ($1, $2);
    my @palette = @{$settings->{_image_palette}} or die "Can't read DDT image with no palette!";

    if ($pixel_format[0] == 4) {
      for my $y (reverse 0..$height-1) {
        for my $x (0..$width-1) {
          $raw_data .= $palette[ ord(substr($data, 2*(1+int($x/4)) + $y*$width*1.5 + $x, 1)) ];
        }
      }
      for my $y (reverse 0..$height-1) {
        for my $x (0..($width/2)-1) {
          $_ = substr($data, 4*(int($x/2)) + $y*$width*1.5 + $x, 1);
          $raw_data .= (chr(vec($_, 1, 4) << 4) x 3) . (chr(vec($_, 0, 4) << 4) x 3);
        }
      }
      $height *= 2;
    } else {
      for my $y (reverse 0..$height-1) {
        for my $x (0..$width-1) {
          $raw_data .= $palette[ ord(substr($data, $y*$width + $x, 1)) ];
        }
      }
    }

  } elsif ($data_format =~ /^DDT_(\d)(7)$/) {
    my @pixel_format = ($1, $2);

    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, $y*$width + $x, 1) x 3;
      }
    }

  } elsif ($data_format =~ /^DDT_(\d)([456])$/) {
    my @pixel_format = ($1, $2);
    my (@colour_pixel, @colour_pixels, $data_byte, $data_pixel, @pixel_values);
    my $pixel_bytes = { 4=>8, 5=>10, 6=>16 }->{$pixel_format[1]};
    my $data_offset = $pixel_format[1]==4 ? 4 : 6;
    my $data_spread = $pixel_format[1]==6 ? 3 : 1;

    for my $y (reverse 0..$height/4-1) {
      @colour_pixels = ();
      for my $x (0..$width/4-1) {
        if ($pixel_format[1] == 4) {
          @colour_pixel = map colour_16_to_bytes($_), unpack S2 => substr($data, $pixel_bytes*($y*$width/4 + $x), 4);
        } elsif ($pixel_format[1] == 5) {
          @colour_pixel = map colour_15_to_bytes($_), unpack S2 => substr($data, $pixel_bytes*($y*$width/4 + $x), 4);
        } elsif ($pixel_format[1] == 6) {
          @colour_pixel = map colour_12_to_bytes($_), unpack S2 => substr($data, $pixel_bytes*($y*$width/4 + $x), 4);
        } else {
          die "Invalid \$pixel_format[1] '$pixel_format[1]'!";
        }
        push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)*2+vec($colour_pixel[1], $_, 8))/3), 0..2;
        push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)+2*vec($colour_pixel[1], $_, 8))/3), 0..2;
        push @colour_pixels, [@colour_pixel];
      }

      for my $sub_y (reverse 0..3) {
        for my $x (0..$width/4-1) {
          $data_byte = substr($data, $data_offset+$data_spread*$sub_y+$pixel_bytes*($y*$width/4 + $x), 1);
          $raw_data .= $colour_pixels[$x][vec($data_byte, $_, 2)] for 0..3;
        }
      }
    }

    if ($pixel_format[0] == 1) { # 1-bit alpha
      my $alpha_byte;
      for my $y (reverse 0..$height/4-1) {
        for my $sub_y (reverse 0..3) {
          for my $x (0..$width/4-1) {
            $alpha_byte = substr($data, 4+10*(int($y*$width/4)+$x), 2);
            $raw_data .= chr(255*vec($alpha_byte, $_+4*$sub_y, 1)) x 3 for 0..3;
          }
        }
      }
      $height *= 2;

    } elsif ($pixel_format[0] == 4) { # 4-bit alpha
      my $alpha_byte;
      for my $y (reverse 0..$height/4-1) {
        for my $sub_y (reverse 0..3) {
          for my $x (0..$width/4-1) {
            $alpha_byte = substr($data, 4+16*(int($y*$width/4)+$x), 12);
            $raw_data .= chr(vec($alpha_byte, $_+6*$sub_y, 4) << 4) x 3 for 0..3;
          }
        }
      }
      $height *= 2;

    }

  } elsif ($data_format eq 'BTX') { # BGRK image, split into RGB and KKK
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= reverse substr($data, ($y*$width + $x)*4, 3);
      }
    }
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= substr($data, ($y*$width + $x)*4+3, 1) x 3;
      }
    }
    $height *= 2;

  } elsif ($data_format eq 'RGB_float_auto') { # RGB, floating-point values, automatically choosing the range
    my ($max, $min, $value);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        for my $v (0..2) {
          $value = unpack f => substr($data, (($y*$width + $x)*3+$v)*4, 4);
          $max = $value if not defined $max or $value > $max;
          $min = $value if not defined $min or $value < $min;
        }
      }
    }
    print "Floating $min -> $max\n";
    my $mult = 255/($max-$min);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= pack C3 => map { ($_-$min) * $mult } unpack f3 => substr($data, ($y*$width + $x)*12, 12);
      }
    }

  } elsif ($data_format eq 'float_auto') { # greyscale floating-point values, automatically choosing the range
    my ($max, $min, $value);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $value = unpack f => substr($data, ($y*$width + $x)*4, 4);
        $max = $value if not defined $max or $value > $max;
        $min = $value if not defined $min or $value < $min;
      }
    }
    print "Floating $min -> $max\n";
    my $mult = 255/($max-$min);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= (pack C => map { ($_-$min) * $mult } unpack f => substr($data, ($y*$width + $x)*4, 4)) x 3;
      }
    }

  } elsif ($data_format =~ /^float_([\d\-\.]+)_([\d\-\.]+)$/) { # greyscale floating-point values
    # Lossy conversion! Not good! (converting to 16-bit PNGs would be better, real floating-point image formats would be best, but paint programs won't often work with them...)
    my ($min, $max) = ($1, $2);
    my $val;
    my $mult = 255/($max-$min);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $val = unpack 'F', substr($data, ($y*$width + $x)*4, 4);
        die "$val out of range $min..$max!" if $val < $min or $val > $max;
        $raw_data .= (pack 'C', ($val-$min)*$mult) x 3;
      }
    }

  } elsif ($data_format eq 'GM1') { # GM1 line data  (extremely uncompleted)
    my @chars = split //, $data;
    my $debug = 1;
    my %counts;
    while (@chars) {
      my $char_c = shift @chars;
      my $char = ord $char_c;

      if ($char == 0x80) {
        $raw_data .= "`";
        ++$counts{'`'};
        ++$counts{px};

      } elsif (0
            or $char_c eq 'D'
            or $char_c eq 'E'
            or $char_c eq 'F'
            or $char_c eq 'G'
            or $char_c eq 'L'
           ) {
        my $new_char = shift @chars;
        $raw_data .= "{$char_c: " if $debug;
        $raw_data .= $new_char;
        $raw_data .= "}" if $debug;

        $counts{px}+=1;

      } elsif ($char_c eq '?'
           ) {
        my $unknown01 = ord(shift @chars);
        my $count = ord(shift @chars);
#        $count %= 66; # ??????????
        $raw_data .= "{$char_c $unknown01 / $count: " if $debug;
#=pod
        if ($unknown01 == 44 and $count == 9) {
#          $count = 12;
        } elsif ($unknown01 == 44 and $count == 66) {
#          $count = 12;
#        } elsif (($unknown01 == 45 or $unknown01 == 46) and $count >= 66) {
        } elsif ($count >= 66) {
          $count = 0;
        }
#=cut

#        $counts{px}+=$count+1;

        $raw_data .= shift @chars for 1..$count+1;
        $raw_data .= "}" if $debug;

      } elsif (0
            or $char_c eq 'A'
            or $char_c eq 'B'
            or $char_c eq 'C'
           ) {
        $raw_data .= "{$char_c: ";
        $raw_data .= shift @chars for 1..2;
        $raw_data .= "}";

        $counts{px}+=2;

      } elsif (0
            or $char_c eq '!'
            or $char_c eq ' '
            or $char_c eq '$'
            or $char_c eq '%'
            or $char_c eq '#'
            or $char_c eq '"'
           ) {
        my $count = ord(shift @chars);
#        $count %= 66; # ??????????
#=pod
        $count = 0 if $count >= 66;
#=cut

        $raw_data .= "{$char_c $count: " if $debug;
        $raw_data .= shift @chars for 1..$count+1;
        $raw_data .= "}" if $debug;

        $counts{px}+=$count+1;

      } else {
        $raw_data .= "{##: " if $debug;
        $raw_data .= $char_c;
        $raw_data .= "}" if $debug;
      }
    }
    $raw_data .= " (" . join(' ', map "$_:$counts{$_}", keys %counts) . ")" if %counts;
    $raw_data .= "\r\n";
  } else {
    die qq{Invalid image data format "$data_format"!\n};
  }

  # then convert into some image format

  my $return = '';
  if ($image_format eq 'BMP') {
    $return = pack("C4LSL4SLSL5", 0x42, 0x4d, 0x66, 0xb4,
                                  0, 0, 0x36, 0x28,
                                  $width, $height, 0x01, 0x18, 0,
                                  $width*$height*3, 0, 0, 0, 0);
                                   #  ^^^^^^^^^^^ probably ought to take account of the padding, but I don't care
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= reverse substr($raw_data, ($y*$width + $x)*3, 3);
      }
      my $pad = $width % 4;
      $return .= pack("C" x $pad, (0) x $pad);
    }

  } elsif ($image_format eq 'raw') {
    $return = $raw_data;

  } else {
    die qq{Invalid image format "$image_format"!\n};
  }
  return $return;
}


# 888, 565, 555 and 444 conversions into 3-byte RGB/BGR
sub colour_32_to_bytes     { return pack C3 => ($_[0] >> 16) & 255, ($_[0] >> 8) & 255, $_[0] & 255 }
sub colour_32_to_bytes_rev { return pack C3 => $_[0] & 255, ($_[0] >> 8) & 255, ($_[0] >> 16) & 255 }
sub colour_16_to_bytes     { return pack C3 => (($_[0] >> 11)) << 3, (($_[0] >> 3) & 252), ($_[0] & 31) << 3 }
sub colour_16_to_bytes_rev { return pack C3 => ($_[0] & 31) << 3, (($_[0] >> 3) & 252), (($_[0] >> 11)) << 3 }
sub colour_15_to_bytes     { return pack C3 => (($_[0] >> 7) & 248), (($_[0] >> 2) & 248), ($_[0] & 31) << 3 }
sub colour_15_to_bytes_rev { return pack C3 => ($_[0] & 31) << 3, (($_[0] >> 2) & 248), (($_[0] >> 7) & 248) }
sub colour_12_to_bytes     { return pack C3 => ($_[0] >> 4) & 240, $_[0] & 240, ($_[0] & 15) << 4 }
sub colour_12_to_bytes_rev { return pack C3 => ($_[0] & 15) << 4, $_[0] & 240, ($_[0] >> 4) & 240 }
# and the other way
sub colour_to_32     { return pack L => (vec($_[0], 0, 8) << 16) | (vec($_[0], 1, 8) << 8) | (vec($_[0], 2, 8)) }
sub colour_to_16     { return pack S => ((vec($_[0], 0, 8) & 248) << 8) | ((vec($_[0], 1, 8) & 252) << 3) | ((vec($_[0], 2, 8) & 248) >> 3) }
sub colour_to_15     { return pack S => ((vec($_[0], 0, 8) & 248) << 7) | ((vec($_[0], 1, 8) & 248) << 2) | ((vec($_[0], 2, 8) & 248) >> 3) }
sub colour_to_12     { return pack S => ((vec($_[0], 0, 8) & 240) << 4) | ((vec($_[0], 1, 8) & 240)) | ((vec($_[0], 2, 8) & 240) >> 4) }
sub colour_to_32_rev { return pack L => (vec($_[0], 2, 8) << 16) | (vec($_[0], 1, 8) << 8) | (vec($_[0], 0, 8)) }
sub colour_to_16_rev { return pack S => ((vec($_[0], 2, 8) & 248) << 8) | ((vec($_[0], 1, 8) & 252) << 3) | ((vec($_[0], 0, 8) & 248) >> 3) }
sub colour_to_15_rev { return pack S => ((vec($_[0], 2, 8) & 248) << 7) | ((vec($_[0], 1, 8) & 248) << 2) | ((vec($_[0], 0, 8) & 248) >> 3) }
sub colour_to_12_rev { return pack S => ((vec($_[0], 2, 8) & 240) << 4) | ((vec($_[0], 1, 8) & 240)) | ((vec($_[0], 0, 8) & 240) >> 4) }

sub limit255 { $_[0] < 0 ? 0 : $_[0] > 255 ? 255 : $_[0] }


sub image_to_data {
  my ($image, $image_format, $data_format) = @_;
  my ($width, $height);
  my $raw_data = '';

  # first convert to a simple RGB format

  if ($image_format eq 'BMP') {
    ($width, $height) = unpack("L2", substr($image, 18, 8));
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $raw_data .= reverse substr($image, 54 + ($y*$width + $x)*3 + $y*($width%4), 3);
      }
    }
  } else {
    die qq{Invalid image format "$image_format"!\n};
  }

  # then convert into some image format

  my $return = '';
  if ($data_format eq 'RGB0') {
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= substr($raw_data, ($y*$width + $x)*3, 3) . "\0";
      }
    }

  } elsif ($data_format eq 'grey') {
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= substr($raw_data, ($y*$width + $x)*3, 1);
      }
    }

  } elsif ($data_format =~ /^grey_x(\d+)$/) {
    my $mult = $1;
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= pack C => (unpack C => substr($raw_data, ($y*$width + $x)*3, 1)) / $mult;
      }
    }

  } elsif ($data_format eq 'BDT') {
    $height /= 2;
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= substr($raw_data, ($y*$width + $x)*3, 1);
        $return .= substr($raw_data, ($height*$width + $y*$width + $x)*3, 1);
      }
    }

  } elsif ($data_format eq 'BTX') {
    $height /= 2;
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= reverse substr($raw_data, ($y*$width + $x)*3, 3);
        $return .= substr($raw_data, ($height*$width + $y*$width + $x)*3, 1);
      }
    }

  } elsif ($data_format eq 'RGB_float_auto') {
    die "Unable to do auto-float-range in image!";

  } elsif ($data_format eq 'float_auto') {
    die "Unable to do auto-float-range in image!";

  } elsif ($data_format =~ /^float_([\d\-\.]+)_([\d\-\.]+)$/) { # greyscale floating-point values
    # Lossy conversion! Not good! (converting to 16-bit PNGs would be better, real floating-point image formats would be best, but paint programs won't work with them...)
    my ($min, $max) = ($1, $2);
    my $mult = 255/($max-$min);
    for my $y (reverse 0..$height-1) {
      for my $x (0..$width-1) {
        $return .= pack 'F', map { ($_ / $mult) + $min } unpack 'C', substr($raw_data, ($y*$width + $x)*3, 1);
      }
    }

  } else {
    die qq{Invalid image data format "$data_format"!\n};
  }
  return $return;
}

sub data_filter_out { # create editable output
  my ($data, $data_format) = @_;
  if ($data_format eq 'scn_trigger') {
    my $return = '';

    require Triggers;

    my @unknowns01 = unpack 'l*' => substr($data, 0, 16, ''); # 9, next-trigger-id, unknown, unknown
die "Invalid triggers_unknown01_0 ($unknowns01[0])!" unless $unknowns01[0] == 9;
    my $trigger_count = unpack l => substr($data, 0, 4, '');

    my @triggers;
    for my $trig_num (0..$trigger_count-1) {
      my %trigger;

      $trigger{unknown01a} = unpack l => substr($data, 0, 4, '');
die "Invalid trigger_unknown01a ($trigger{unknown01a})!" unless $trigger{unknown01a} == 5;

      $trigger{id} = unpack l => substr($data, 0, 4, '');

      $trigger{group} = unpack l => substr($data, 0, 4, '');

      $trigger{priority} = unpack l => substr($data, 0, 4, '');
die "Invalid priority $trigger{priority}!\n" unless $Triggers::priorities_to_text[$trigger{priority}];

      $trigger{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');

      $trigger{eventid} = unpack l => substr($data, 0, 4, '');

      $trigger{loop} = unpack c => substr($data, 0, 1, '');
      $trigger{active} = unpack c => substr($data, 0, 1, '');
      $trigger{run_immediately} = unpack c => substr($data, 0, 1, '');

      my $condition_count = unpack l => substr($data, 0, 4, '');
      for my $cond_num (0..$condition_count-1) {
        my %condition;
        $condition{unknown01} = unpack l => substr($data, 0, 4, '');
die "Invalid condition unknown01 '$condition{unknown01}'!" unless $condition{unknown01} == 2;

        $condition{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
        $condition{text02} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
warn "WARNING: Condition text01 ne text02! ('$condition{text01}' ne '$condition{text02}')" if $condition{text01} ne $condition{text02};

        my $param_count = unpack l => substr($data, 0, 4, '');
        for my $param_num (0..$param_count-1) {
          my %parameter;
          $parameter{unknown01} = unpack l => substr($data, 0, 4, '');
die "Invalid param unknown01 '$parameter{unknown01}'!" unless $parameter{unknown01} == 4;

          $parameter{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $parameter{text02} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $parameter{var_type} = unpack l => substr($data, 0, 4, '');

# my ($match) = grep $_->[0] eq $parameter{text01},
# @{$Triggers::condition_from_text{$Triggers::condition_to_text{$condition{text01}}[0]}[1][1]};
# ++$main::triggertypes{"$match->[3] => $parameter{var_type}"};

          my $value_count = unpack l => substr($data, 0, 4, '');
          $parameter{string_id} = unpack l => substr($data, 0, 4, '')
            if $parameter{var_type} == 22;

          $parameter{values} = [
            map Convert::utf16to8(substr($data, 0, 2*unpack(l => substr($data, 0, 4, '')), '')), 0..$value_count-1
          ];
#          @values = map $triggers[
#          $parameter{value} = join ';', map Convert::utf16to8($_), @values;
#          $parameter{values} = [
#            map Convert::utf16to8(substr($data, 0, 2*unpack(l => substr($data, 0, 4, '')), '')),
#              0..$value_count-1
#          ];
          push @{$condition{params}}, \%parameter;
        }

        $condition{check_code} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');

        my $code_count = unpack l => substr($data, 0, 4, '');
        for my $code_num (0..$code_count-1) {
          my %code;
          $code{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $code{looping} = unpack c => substr($data, 0, 1, '');
          $code{unknown01} = unpack l => substr($data, 0, 4, '');
die "Code unknown01 != looping ($code{unknown01} != $code{looping}!" if $code{unknown01} != $code{looping};
          if ($code{looping}) {
            $code{loop_param} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          }
          push @{$condition{codes}}, \%code;
        }

        push @{$trigger{conditions}}, \%condition;
      }

      my $effect_count = unpack l => substr($data, 0, 4, '');
      for my $cond_num (0..$effect_count-1) {
        my %effect;
        $effect{unknown01} = unpack l => substr($data, 0, 4, '');
die "Invalid effect unknown01 '$effect{unknown01}'!" unless $effect{unknown01} == 2;
        $effect{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
        $effect{text02} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
warn "WARNING: Effect text01 ne text02! ('$effect{text01}' ne '$effect{text02}')" if $effect{text01} ne $effect{text02};

        my $param_count = unpack l => substr($data, 0, 4, '');
        for my $param_num (0..$param_count-1) {
          my %parameter;
          $parameter{unknown01} = unpack l => substr($data, 0, 4, '');
die "Invalid param unknown01 '$parameter{unknown01}'!" unless $parameter{unknown01} == 4;
          $parameter{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $parameter{text02} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $parameter{var_type} = unpack l => substr($data, 0, 4, '');

# my ($match) = grep $_->[0] eq $parameter{text01},
# @{$Triggers::effect_from_text{$Triggers::effect_to_text{$effect{text01}}[0]}[1][1]};
# ++$main::triggertypes{"$match->[3] => $parameter{var_type}"};

          my $value_count = unpack l => substr($data, 0, 4, '');
          $parameter{string_id} = unpack l => substr($data, 0, 4, '')
            if $parameter{var_type} == 22;

          $parameter{values} = [
            map Convert::utf16to8(substr($data, 0, 2*unpack(l => substr($data, 0, 4, '')), '')), 0..$value_count-1
          ];
#          $parameter{value} = join ';', 
##          $parameter{values} = [
#            map Convert::utf16to8(substr($data, 0, 2*unpack(l => substr($data, 0, 4, '')), '')),
#              0..$value_count-1
#          ]
#          ;
          push @{$effect{params}}, \%parameter;
        }

        $effect{text03} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
die "Invalid effect_text03 '$effect{text03}'!" unless $effect{text03} eq 'true';

        my $code_count = unpack l => substr($data, 0, 4, '');
        for my $code_num (0..$code_count-1) {
          my %code;
          $code{text01} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          $code{looping} = unpack c => substr($data, 0, 1, '');
          $code{unknown01} = unpack l => substr($data, 0, 4, '');
die "Code unknown01 != looping ($code{unknown01} != $code{looping}!" if $code{unknown01} != $code{looping};
          if ($code{looping}) {
            $code{loop_param} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');
          }
          push @{$effect{codes}}, \%code;
        }

        push @{$trigger{effects}}, \%effect;
      }

      $triggers[$trigger{id}] = \%trigger;
    }

    my $group_count = unpack l => substr($data, 0, 4, '');
    my @groups;
    for my $group_num (0..$group_count-1) {
      my %group;

      $group{unknown01} = unpack l => substr($data, 0, 4, '');
die "Invalid unknown01 != 1!" if $group{unknown01} != 1;

      $group{id} = unpack l => substr($data, 0, 4, '');
      $group{name} = unpack 'Z*' => substr($data, 0, unpack(l => substr($data, 0, 4, '')), '');

      my $content_count = unpack l => substr($data, 0, 4, '');
      for (0..$content_count-1) {
        push @{$group{items}}, unpack l => substr($data, 0, 4, '');
      }
      push @groups, \%group;
    }

#print join "\n", sort keys %main::triggertypes;

#############
#
######

    my $indent = "\t";

#    $return .= qq{<?xml version="1.0" encoding="UTF-8"?>\n\n<Triggers>\n};
#    $return .= qq{$indent<!-- unknowns: }.join(', ', @unknowns01).qq{ -->\n};
    $return .= qq{<?xml version="1.0" encoding="UTF-8"?>\n\n<Triggers Unknowns="}.join(', ', @unknowns01[2,3]).qq{">\n};

    foreach my $group (@groups) {
      $return .= qq{$indent<Group Name="}.Convert::xml_encode_lesser($group->{name}).qq{">\n};
#$return .= qq{<!-- unknown: $group->{unknown01} -->\n};

      foreach (@{$group->{items}}) {
        my $indent = "$indent\t";
        my $trigger = $triggers[$_];
        $return .= qq{$indent<Trigger Name="}.Convert::xml_encode_lesser($trigger->{text01}).qq{" Priority="$Triggers::priorities_to_text[$trigger->{priority}]" Loop="$Triggers::yesno_to_text[$trigger->{loop}]" RunImmediately="$Triggers::yesno_to_text[$trigger->{run_immediately}]" Active="$Triggers::yesno_to_text[$trigger->{active}]"}
                . ($trigger->{eventid} != -1 ? qq{ EventID="$trigger->{eventid}"}:'')
                . qq{>\n};
#$return .= qq{<!-- unknown: $trigger->{unknown02} -->\n};

        foreach my $condition (@{$trigger->{conditions}}) {
          my $indent = "$indent\t";
          die "Unrecognised condition name '$condition->{text01}'!" unless exists $Triggers::condition_to_text{$condition->{text01}};
          $return .= qq{$indent<Condition }
                  . ($condition->{text01} ne $condition->{text02} ? qq{Name="}.Convert::xml_encode_lesser($condition->{text02}).qq{" }:'')
                  . qq{Type="$Triggers::condition_to_text{$condition->{text01}}[0]"};
          foreach my $param (@{$condition->{params}}) {
            my @values = @{$param->{values}};

            if ($param->{var_type} == 3) {
              @values = map ucfirst, @values; # true -> True, etc
            } elsif ($param->{var_type} == 8) {
              @values = map $_==-1?'(none)':$triggers[$_]?$triggers[$_]{text01}:'(none)', @values;
            } elsif ($param->{var_type} == 11) {
              @values = map +($Triggers::techstatuses_to_text[$_]||die"Invalid tech status '$_'!"), @values;
            } elsif ($param->{var_type} == 15) {
              @values = map +($Triggers::ops_to_text{$_}||die"Invalid operator '$_'!"), @values;
            } elsif ($param->{var_type} == 17) {
              @values = map /^(\d+),(\d+)$/?"Player $1, Army ".($2+1):$_, @values;
            } elsif ($param->{var_type} == 22) {
              @values = map "\$\$$param->{string_id}\$\$$_", @values;
            }

            $return .= qq{ $Triggers::condition_to_text{$condition->{text01}}[1]{$param->{text01}}="}
                    . join(';', map Convert::xml_encode_lesser($_), @values)
                    . qq{"};
# $return .= qq{<!--type=$Triggers::vartypes_to_text[$param->{var_type}]} . ($param->{var_type}==22 ? qq{;u2=$param->{unknown02}} : '') . qq{-->};
          }
          $return .= qq{/>\n};
        }

        foreach my $effect (@{$trigger->{effects}}) {
          my $indent = "$indent\t";
          if ($effect->{text01} eq 'XSCode') {
            $return .= qq{$indent<Code>}
                    . Convert::xml_encode_muchlesser($effect->{codes}[0]{text01})
                    . qq{</Code>\n};

          } else {
            die "Unrecognised effect name '$effect->{text01}'!" unless exists $Triggers::effect_to_text{$effect->{text01}};
            $return .= qq{$indent<Effect }
                    . ($effect->{text01} ne $effect->{text02} ? qq{Name="}.Convert::xml_encode_lesser($effect->{text02}).qq{" }:'')
                    . qq{Type="$Triggers::effect_to_text{$effect->{text01}}[0]"};
            foreach my $param (@{$effect->{params}}) {
              my @values = @{$param->{values}};
  
              if ($param->{var_type} == 3) {
                @values = map ucfirst, @values; # true -> True, etc
              } elsif ($param->{var_type} == 8) {
                @values = map $_==-1?'(none)':$triggers[$_]?$triggers[$_]{text01}:'(none)', @values;
              } elsif ($param->{var_type} == 11) {
                @values = map +($Triggers::techstatuses_to_text[$_]||die"Invalid tech status '$_'!"), @values;
              } elsif ($param->{var_type} == 15) {
                @values = map +($Triggers::ops_to_text{$_}||die"Invalid operator '$_'!"), @values;
              } elsif ($param->{var_type} == 17) {
                @values = map /^(\d+),(\d+)$/?"Player $1, Army ".($2+1):$_, @values;
              } elsif ($param->{var_type} == 22) {
                @values = map "\$\$$param->{string_id}\$\$$_", @values;
              }
  
              $return .= qq{ $Triggers::effect_to_text{$effect->{text01}}[1]{$param->{text01}}="}
                      . join(';', map Convert::xml_encode_lesser($_), @values)
                      . qq{"};
  # $return .= qq{<!--type=$var_types[$param->{var_type}]-->};
            }
  
            # Nasty hack (for e.g. cine04_2)
            if ($effect->{text01} eq 'BlockAllSounds' and not grep $_->{text01} eq 'Dialog', @{$effect->{params}}) {
              $return .= ' ExcludeDialog="True"';
              warn qq{Guessing ExcludeDialog="True" for 'BlockAllSounds' effect\n};
            }
            if ($effect->{text01} eq 'Move To Point' and not grep $_->{text01} eq 'AttackMove', @{$effect->{params}}) {
              $return .= ' AttackMove="False"';
              warn qq{Guessing AttackMove="False" for 'Move To Point' effect\n};
            }
  
            $return .= qq{/>\n};
          }
        }
        $return .= qq{$indent</Trigger>\n};
      }
      $return .= qq{$indent</Group>\n};
    }
    $return .= qq{</Triggers>};

    $return =~ s/\n/\r\n/g; # since it's writing in binmode
    return $return;

  } else {
    die "Invalid data format '$data_format'!";
  }
}

sub data_filter_in {
  my ($data, $data_format) = @_;

  if ($data_format eq 'scn_trigger') {
    my $return = '';

    require Triggers;

    (undef,undef,undef,undef)=($Triggers::effect_from_text,
                               $Triggers::yesno_from_text,
                               $Triggers::condition_from_text,
                               $Triggers::priorities_from_text); # avoid "Name ... used only once" warnings

    $data =~ s/\r\n/\n/g;

    require XML::Parser;

    my $parser = XML::Parser->new(Style => 'Tree', ErrorContext => 2);
    my $xml_tree;
    eval { $xml_tree = $parser->parse($data) };
    die qq{Error in XML code:$@\n} if $@;

    $xml_tree->[0] eq 'Triggers' or die "Top level tag must be 'Triggers', not '$xml_tree->[0]'!";
    my ($triggers_attrs, @triggers_contents) = @{$xml_tree->[1]};
    my @unknowns01 = split /,\s+/, $triggers_attrs->{Unknowns} or die "Can't find Unknowns in Triggers!";

    foreach my $p (keys %$triggers_attrs) {
      die "Unrecognised parameter '$p' in <Triggers>!" unless grep $p eq $_,
        qw(Uneditable Unknowns);
    }

    my $uneditable = (defined $triggers_attrs->{Uneditable} and $triggers_attrs->{Uneditable} eq 'True') ? 1 : 0;

    my %trigger_ids = ('(none)' => -1);

    my (@triggers, @groups);

    for (0..@triggers_contents/2-1) {
      next if $triggers_contents[$_*2] eq '0';
      $triggers_contents[$_*2] eq 'Group' or die "<Triggers> can only contain <Group> tags, not '$triggers_contents[$_*2]'!";

      my ($group_attrs, @group_contents) = @{$triggers_contents[$_*2+1]};

      foreach my $p (keys %$group_attrs) {
        die "Unrecognised parameter '$p' in <Group>!" unless grep $p eq $_,
          qw(Name);
      }

      push @groups, {
        id => scalar @groups,
        name => $group_attrs->{Name},
        trigger_ids => [],
      };

      for (0..@group_contents/2-1) {
        next if $group_contents[$_*2] eq '0';
        $group_contents[$_*2] eq 'Trigger' or die "<Group> tags can only contain <Trigger> tags, not '$group_contents[$_*2]'!";

        my ($trigger_attrs, @trigger_contents) = @{$group_contents[$_*2+1]};

        foreach my $p (keys %$trigger_attrs) {
          die "Unrecognised parameter '$p' in <Trigger>!" unless grep $p eq $_,
            qw(Name Priority Loop RunImmediately Active EventID);
        }

        push @triggers, {
          id => scalar @triggers,
          group => $#groups,
          name => $trigger_attrs->{Name},
          priority => $trigger_attrs->{Priority},
          loop => $trigger_attrs->{Loop},
          run_immediately => $trigger_attrs->{RunImmediately},
          active => $trigger_attrs->{Active},
          eventid => (exists $trigger_attrs->{EventID} ? $trigger_attrs->{EventID} : -1),
          conditions => [],
          effects => [],
        };
        die "Duplicate trigger name '$triggers[-1]{name}'!" if exists $trigger_ids{$triggers[-1]{name}};
        $trigger_ids{$triggers[-1]{name}} = $#triggers;
        push @{$groups[-1]{trigger_ids}}, $#triggers;

        for (0..@trigger_contents/2-1) {
          next if $trigger_contents[$_*2] eq '0';

          my ($part_attrs, @part_contents) = @{$trigger_contents[$_*2+1]};
          if ($trigger_contents[$_*2] eq 'Condition') {
            push @{$triggers[-1]{conditions}}, $part_attrs;
          } elsif ($trigger_contents[$_*2] eq 'Effect') {
            push @{$triggers[-1]{effects}}, $part_attrs;
          } elsif ($trigger_contents[$_*2] eq 'Code') {
            push @{$triggers[-1]{effects}}, [$part_contents[1]];
          } else {
            die "'Trigger' tags can only contain 'Condition' and 'Effect' tags, not '$trigger_contents[$_*2]'!";
          }
        }
      }
    }

###

    $return .= pack l => 9;
    if ($uneditable) {
      $return .= pack l => -2147483648;
    } else {
      $return .= pack l => scalar @triggers;
    }
    $return .= pack 'l*' => @unknowns01;
    $return .= pack l => scalar @triggers;
    foreach my $trigger (@triggers) {
      $return .= pack l => 5;
      $return .= pack 'l*' => $trigger->{id}, $trigger->{group}, $Triggers::priorities_from_text{$trigger->{priority}};
      if ($uneditable) {
        $return .= pack 'l/Z*' => trigger_id_to_unique_text($trigger->{id});
      } else {
        $return .= pack 'l/Z*' => $trigger->{name};
      }
      $return .= pack l => $trigger->{eventid};
      $return .= pack 'c*' => map $Triggers::yesno_from_text{$trigger->{$_}}, qw(loop active run_immediately);

      $return .= pack l => scalar @{$trigger->{conditions}};
      foreach my $cond (@{$trigger->{conditions}}) {
        my $condition = $Triggers::condition_from_text{$cond->{Type}};
        $return .= pack l => 2;

        if ($uneditable) {
          $return .= (pack 'l/Z*' => '') x 2;
        } else {
          $return .= pack 'l/Z*' => $condition->[0];
          $return .= pack 'l/Z*' => exists($cond->{Name})?$cond->{Name}:$condition->[0];
        }

        $return .= pack l => scalar @{$condition->[1][1]};
        foreach my $param (@{$condition->[1][1]}) {
          $return .= pack l => 4;
          if ($uneditable) {
            $return .= pack 'l/Z*l/Z*' => $param->[0], '';
          } else {
            $return .= pack 'l/Z*l/Z*' => $param->[0], $param->[2];
          }
          $return .= pack l => $Triggers::vartype_from_text{$param->[3]};

          die "Condition '$cond->{Type}' requires a '$param->[1]' parameter!" unless exists $cond->{$param->[1]};

          my @values;
          if ($param->[3] eq 'bool') {
            @values = (lcfirst $cond->{$param->[1]});
            $return .= pack l => 1;
          } elsif ($param->[3] eq 'event') {
            die "Unrecognised trigger name '$cond->{$param->[1]}'!" unless exists $trigger_ids{$cond->{$param->[1]}};
            @values = ($trigger_ids{$cond->{$param->[1]}});
            $return .= pack l => 1;
          } elsif ($param->[3] eq 'techstatus') {
            @values = $Triggers::techstatuses_from_text{$cond->{$param->[1]}};
            $return .= pack l => 1;
          } elsif ($param->[3] eq 'operator') {
            @values = $Triggers::ops_from_text{$cond->{$param->[1]}};
            $return .= pack l => 1;
          } elsif ($param->[3] eq 'stringid') {
            $cond->{$param->[1]} =~ /^\$\$(\d+)\$\$(.*)$/ or die "Invalid stringid value - must be '\$\$number\$\$text'!";
            @values = ($2);
            $return .= pack 'l*' => 1, $1;
          } elsif ($param->[3] eq 'sound' or $param->[3] eq 'unit') {
            @values = split ';', $cond->{$param->[1]};
            $return .= pack l => scalar @values;
          } elsif ($param->[3] eq 'group') {
            @values = ($cond->{$param->[1]});
            @values = ("$1,".($2-1)) if $cond->{$param->[1]} =~ /Player\s*(\d+)\s*,\s*Army\s*(\d+)/i;
            $return .= pack l => 1;
          } else {
            @values = ($cond->{$param->[1]});
            $return .= pack l => 1;
          }
          foreach (map Convert::utf8to16($_), @values) {
            $return .= pack 'la*' => length($_)/2, $_;
          }
        }

        $return .= pack 'l/Z*' => $condition->[1][2][0];
        $return .= pack l => @{$condition->[1][3]}/2;
        for (0..@{$condition->[1][3]}/2-1) {
          $return .= pack 'l/Z*' => $condition->[1][3][$_*2];
          $return .= pack c => $condition->[1][3][$_*2+1] ? 1 : 0;
          $return .= pack l => $condition->[1][3][$_*2+1] ? 1 : 0;
          $return .= pack 'l/Z*' => $condition->[1][3][$_*2+1] if $condition->[1][3][$_*2+1];
        }

      }

      $return .= pack l => scalar @{$trigger->{effects}};
      foreach my $eff (@{$trigger->{effects}}) {

        if (ref($eff) eq 'ARRAY') {
          # <Code> statement

          $return .= pack l => 2;
          if ($uneditable) {
            $return .= (pack 'l/Z*' => '') x 2;
          } else {
            $return .= (pack 'l/Z*' => 'XSCode') x 2;
          }
          $return .= pack l => 0;
          $return .= pack 'l/Z*' => 'true';
          $return .= pack l => 1;

          $return .= pack 'l/Z*' => $eff->[0];
          $return .= pack cl => 0, 0;

        } else {
          # standard effect

          my $effect = $Triggers::effect_from_text{$eff->{Type}};
          $return .= pack l => 2;
  
          if ($uneditable) {
            $return .= (pack 'l/Z*' => '') x 2;
          } else {
            $return .= pack 'l/Z*' => $effect->[0];
            $return .= pack 'l/Z*' => exists($eff->{Name})?$eff->{Name}:$effect->[0];
          }
  
          $return .= pack l => scalar @{$effect->[1][1]};
          foreach my $param (@{$effect->[1][1]}) {
            $return .= pack l => 4;
            if ($uneditable) {
              $return .= pack 'l/Z*l/Z*' => $param->[0], '';
            } else {
              $return .= pack 'l/Z*l/Z*' => $param->[0], $param->[2];
            }
            $return .= pack l => $Triggers::vartype_from_text{$param->[3]};
  
            die "Effect '$eff->{Type}' requires a '$param->[1]' parameter!" unless exists $eff->{$param->[1]};
  
            my @values;
            if ($param->[3] eq 'bool') {
              @values = (lcfirst $eff->{$param->[1]});
              $return .= pack l => 1;
            } elsif ($param->[3] eq 'event') {
              die "Unrecognised trigger name '$eff->{$param->[1]}'!" unless exists $trigger_ids{$eff->{$param->[1]}};
              @values = ($trigger_ids{$eff->{$param->[1]}});
              $return .= pack l => 1;
            } elsif ($param->[3] eq 'techstatus') {
              @values = $Triggers::techstatuses_from_text{$eff->{$param->[1]}};
              $return .= pack l => 1;
            } elsif ($param->[3] eq 'operator') {
              @values = $Triggers::ops_from_text{$eff->{$param->[1]}};
              $return .= pack l => 1;
            } elsif ($param->[3] eq 'stringid') {
              $eff->{$param->[1]} =~ /^\$\$(\d+)\$\$(.*)$/ or die "Invalid stringid value - must be '\$\$number\$\$text'!";
              @values = ($2);
              $return .= pack 'l*' => 1, $1;
            } elsif ($param->[3] eq 'sound' or $param->[3] eq 'unit') {
              @values = split ';', $eff->{$param->[1]};
              $return .= pack l => scalar @values;
            } elsif ($uneditable and $param->[1] eq 'TriggerName' and $param->[3] eq 'string') {
              @values = (trigger_id_to_unique_text($trigger_ids{$eff->{$param->[1]}}));
              $return .= pack l => 1;
            } elsif ($param->[3] eq 'group') {
              @values = ($eff->{$param->[1]});
              @values = ("$1,".($2-1)) if $eff->{$param->[1]} =~ /Player\s*(\d+)\s*,\s*Army\s*(\d+)/i;
              $return .= pack l => 1;
            } else {
              @values = ($eff->{$param->[1]});
              $return .= pack l => 1;
            }
            foreach (map Convert::utf8to16($_), @values) {
              $return .= pack 'la*' => length($_)/2, $_;
            }
          }
  
          $return .= pack 'l/Z*' => 'true';
          $return .= pack l => @{$effect->[1][3]}/2;
          for (0..@{$effect->[1][3]}/2-1) {
            $return .= pack 'l/Z*' => $effect->[1][3][$_*2];
            $return .= pack c => $effect->[1][3][$_*2+1] ? 1 : 0;
            $return .= pack l => $effect->[1][3][$_*2+1] ? 1 : 0;
            $return .= pack 'l/Z*' => $effect->[1][3][$_*2+1] if $effect->[1][3][$_*2+1];
          }
        }
      }
    }

    $return .= pack l => scalar @groups;
    foreach my $group (@groups) {
      $return .= pack l => 1;
      $return .= pack l => $group->{id};
      $return .= pack 'l/Z*' => $group->{name};
      $return .= pack l => scalar @{$group->{trigger_ids}};
      $return .= pack 'l*' => @{$group->{trigger_ids}};
    }

    return $return;

  } else {
    die "Invalid data format '$data_format'!";
  }
}

sub trigger_id_to_unique_text {
  return "ProtectedByAOMEd" if $_[0] == 0;
  my $id = pack S => $_[0];
  my @codes = qw(_ 0 1 O I l t f j x y z v r w q);
  return join '', map $codes[vec($id, $_, 4)], reverse 0..3;
}

sub pack_f { # works with 00000080 <-> -0 (instead of 0)
  if ($_[0] eq '-0') {
    "\0\0\0\x80";
  } else {
    pack f => $_[0];
  }
}
sub unpack_f {
  if ($_[0] eq "\0\0\0\x80") {
    '-0';
  } else {
    unpack f => $_[0];
  }
}



=pod
sub c{my$a=shift;$a=~s/[^\x20-\x7F]/_/g;$a;''}
sub bzone_decompress {
  my ($data, $format) = @_;

  use constant BZ_DEBUG => 0;

  # mostly copied from lzo1x_d.ch

  use constant M2_MAX_OFFSET => 0x0800;

#  my $prefix = substr $data, 0, 1, '';
  my $output = '';
  my $pos = 0;
  my ($t, $m_pos);

  if (ord(substr $data, 0, 1) > 17) {
print ">17\n" if BZ_DEBUG;
    $t = ord(substr $data, $pos++, 1) - 17;
    if ($t < 4) {
      goto match_next;
    }
    $output .= substr $data, $pos, $t; $pos += $t;
print '0) ',c($output),"\n" if BZ_DEBUG;
    goto first_literal_run;
  }

  while ($pos < length $data) {
    $t = ord(substr $data, $pos++, 1);
    if ($t >= 16) {
      goto match;
    }
    if ($t == 0) {
      while (ord(substr $data, $pos, 1) == 0) {
        $t += 255;
        ++$pos;
      }
      $t += 15 + ord(substr $data, $pos++, 1);
    }
#    $output .= substr $data, $pos, 3; $pos += 3;
#    if (--$t > 0) {
#      $output .= substr $data, $pos, $t; $pos += $t;
#    }
    $output .= substr $data, $pos, $t+3; $pos += $t+3;
print '1) ',c($output),"\n" if BZ_DEBUG;

first_literal_run:
   
    $t = ord(substr $data, $pos++, 1);
    if ($t >= 16) {
      goto match;
    }
    $m_pos = length($output) - (1 + M2_MAX_OFFSET);
    $m_pos -= $t >> 2;
    $m_pos -= ord(substr $data, $pos++, 1) << 2;
    $output .= substr $output, $m_pos, 3;
print 'f) ',c($output),"\n" if BZ_DEBUG;
    goto match_done;

    while ($pos < length $data) {
match:
      if ($t >= 64) {
print "M2 [$t]\n" if BZ_DEBUG;
        $m_pos = length($output)-1;

#        # assume it's LZO1X
#        $m_pos -= ($t >> 2) & 7;
#        $m_pos -= ord(substr $data, $pos++, 1) << 3;
#        $t = ($t >> 5) - 1;
        # assume it's LZO1Y
        $m_pos -= ($t >> 2) & 3;
        $m_pos -= ord(substr $data, $pos++, 1) << 2;
        $t = ($t >> 4) - 3;

        goto copy_match;
      } elsif ($t >= 32) {
print "M3\n" if BZ_DEBUG;
        $t &= 31;
        if ($t == 0) {
          while (ord(substr $data, $pos, 1) == 0) {
            $t += 255;
            ++$pos;
          }
          $t += 31 + ord(substr $data, $pos++, 1);
        }
        $m_pos = length($output)-1;
        $m_pos -= (ord(substr $data, $pos, 1) >> 2) + (ord(substr $data, $pos+1, 1) << 6);
        $pos += 2;
      } elsif ($t >= 16) {
print "M4\n" if BZ_DEBUG;
        $m_pos = length($output);
        $m_pos -= ($t & 8) << 11;
        $t &= 7;
        if ($t == 0) {
          while (ord(substr $data, $pos, 1) == 0) {
            $t += 255;
            ++$pos;
          }
          $t += 7 + ord(substr $data, $pos++, 1);
        }
        $m_pos -= (ord(substr $data, $pos, 1) >> 2) + (ord(substr $data, $pos+1, 1) << 6);
        $pos += 2;
        if ($m_pos == length($output)) {
          goto eof_found;
        }
        $m_pos -= 0x4000;
      } else {
print "M1\n" if BZ_DEBUG;
        $m_pos = length($output)-1;
        $m_pos -= $t >> 2;
        $m_pos -= ord(substr $data, $pos++, 1) << 2;
        $output .= substr $output, $m_pos, 2;
        goto match_done;
      }

copy_match:
print "copy_match $m_pos, $t\n" if BZ_DEBUG;
      $output .= substr $output, $m_pos, $t+2;

print 'c) ',c($output),"\n" if BZ_DEBUG;
match_done:
      $t = ord(substr $data, $pos-2, 1) & 3;
      if ($t == 0) {
        last;
      }
match_next:
      $output .= substr $data, $pos, $t; $pos += $t;
print 'n) ',c($output),"\n" if BZ_DEBUG;
select undef,undef,undef,0.4 if BZ_DEBUG;
      $t = ord(substr $data, $pos++, 1);
    }
  }
eof_found:
  if (BZ_DEBUG) {
    print ">>>\n$output";
    exit;
  }
  return $output;
}
=cut

1; # to please wrapper.pl
